#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module inherits constants for Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import re
from collections import namedtuple
from typing import Literal

import attr

# Attr constants
DENON_ATTR_SETATTR = [attr.setters.validate, attr.setters.convert]

# Types for constants
ReceiverType = namedtuple("ReceiverType", ["type", "port"])
DescriptionType = namedtuple("DescriptionType", ["port", "url"])
ReceiverURLs = namedtuple(
    "ReceiverURLs",
    [
        "appcommand",
        "appcommand0300",
        "status",
        "mainzone",
        "deviceinfo",
        "netaudiostatus",
        "tunerstatus",
        "hdtunerstatus",
        "command_sel_src",
        "command_fav_src",
        "command_power_on",
        "command_power_standby",
        "command_volume_up",
        "command_volume_down",
        "command_set_volume",
        "command_mute_on",
        "command_mute_off",
        "command_sel_sound_mode",
        "command_neural_x_on_off",
        "command_imax_auto_off",
        "command_imax_audio_settings",
        "command_imax_hpf",
        "command_imax_lpf",
        "command_imax_subwoofer_mode",
        "command_imax_subwoofer_output",
        "command_cinema_eq",
        "command_netaudio_post",
        "command_set_all_zone_stereo",
        "command_pause",
        "command_play",
        "command_cusor_up",
        "command_cusor_down",
        "command_cusor_left",
        "command_cusor_right",
        "command_cusor_enter",
        "command_back",
        "command_info",
        "command_options",
        "command_setup_open",
        "command_setup_close",
        "command_setup_query",
        "command_channel_level_adjust",
        "command_dimmer_toggle",
        "command_dimmer_set",
        "command_channel_volume",
        "command_channel_volumes_reset",
        "command_subwoofer_on_off",
        "command_subwoofer_level",
        "command_lfe",
        "command_tactile_transducer",
        "command_tactile_transducer_level",
        "command_tactile_transducer_lpf",
        "command_delay_up",
        "command_delay_down",
        "command_auromatic_3d_preset",
        "command_auromatic_3d_strength",
        "command_auro_3d_mode",
        "command_dirac_filter",
        "command_eco_mode",
        "command_lfc",
        "command_containment_amount",
        "command_loudness_management",
        "command_bass_sync",
        "command_dialog_enhancer",
        "command_hdmi_output",
        "command_hdmi_audio_decode",
        "command_quick_select_mode",
        "command_quick_select_memory",
        "command_smart_select_mode",  # Marantz
        "command_smart_select_memory",  # Marantz
        "command_auto_standby",
        "command_sleep",
        "command_center_spread",
        "command_video_processing_mode",
        "command_room_size",
        "command_status",
        "command_system_reset",
        "command_network_restart",
        "command_trigger",
        "command_speaker_preset",
        "command_bluetooth_transmitter",
        "command_dialog_control",
        "command_speaker_virtualizer",
        "command_effect_speaker_selection",
        "command_drc",
        "command_delay_time",
        "command_audio_restorer",
        "command_remote_control_lock",
        "command_panel_lock",
        "command_panel_and_volume_lock",
        "command_graphic_eq",
        "command_headphone_eq",
    ],
)
TelnetCommands = namedtuple(
    "TelnetCommands",
    [
        "command_sel_src",
        "command_fav_src",
        "command_power_on",
        "command_power_standby",
        "command_volume_up",
        "command_volume_down",
        "command_set_volume",
        "command_mute_on",
        "command_mute_off",
        "command_sel_sound_mode",
        "command_neural_x_on_off",
        "command_imax_auto_off",
        "command_imax_audio_settings",
        "command_imax_hpf",
        "command_imax_lpf",
        "command_imax_subwoofer_mode",
        "command_imax_subwoofer_output",
        "command_cinema_eq",
        "command_set_all_zone_stereo",
        "command_pause",
        "command_play",
        "command_multieq",
        "command_dynamiceq",
        "command_reflevoffset",
        "command_dynamicvol",
        "command_tonecontrol",
        "command_bass",
        "command_treble",
        "command_cusor_up",
        "command_cusor_down",
        "command_cusor_left",
        "command_cusor_right",
        "command_cusor_enter",
        "command_back",
        "command_info",
        "command_options",
        "command_setup_open",
        "command_setup_close",
        "command_setup_query",
        "command_channel_level_adjust",
        "command_dimmer_toggle",
        "command_dimmer_set",
        "command_channel_volume",
        "command_channel_volumes_reset",
        "command_subwoofer_on_off",
        "command_subwoofer_level",
        "command_lfe",
        "command_tactile_transducer",
        "command_tactile_transducer_level",
        "command_tactile_transducer_lpf",
        "command_delay_up",
        "command_delay_down",
        "command_auromatic_3d_preset",
        "command_auromatic_3d_strength",
        "command_auro_3d_mode",
        "command_dirac_filter",
        "command_eco_mode",
        "command_lfc",
        "command_containment_amount",
        "command_loudness_management",
        "command_bass_sync",
        "command_dialog_enhancer",
        "command_hdmi_output",
        "command_hdmi_audio_decode",
        "command_quick_select_mode",
        "command_quick_select_memory",
        "command_smart_select_mode",  # Marantz
        "command_smart_select_memory",  # Marantz
        "command_auto_standby",
        "command_sleep",
        "command_center_spread",
        "command_video_processing_mode",
        "command_room_size",
        "command_status",
        "command_system_reset",
        "command_network_restart",
        "command_trigger",
        "command_speaker_preset",
        "command_bluetooth_transmitter",
        "command_dialog_control",
        "command_speaker_virtualizer",
        "command_effect_speaker_selection",
        "command_drc",
        "command_delay_time",
        "command_audio_restorer",
        "command_remote_control_lock",
        "command_panel_lock",
        "command_panel_and_volume_lock",
        "command_graphic_eq",
        "command_headphone_eq",
    ],
)

# AVR-X search patterns
DEVICEINFO_AVR_X_PATTERN = re.compile(
    r"(.*AV(C|R)-(X|S|A).*|.*SR500[6-9]|.*SR60(07|08|09|10|11|12|13)|."
    r"*SR70(07|08|09|10|11|12|13)|.*SR501[3-4]|.*NR1604|.*NR1710)"
)

DEVICEINFO_COMMAPI_PATTERN = re.compile(r"(0210|0220|0250|0300|0301)")

# Receiver types
AVR_NAME = "avr"
AVR_X_NAME = "avr-x"
AVR_X_2016_NAME = "avr-x-2016"

AVR = ReceiverType(type=AVR_NAME, port=80)
AVR_X = ReceiverType(type=AVR_X_NAME, port=80)
AVR_X_2016 = ReceiverType(type=AVR_X_2016_NAME, port=8080)
VALID_RECEIVER_TYPES = (AVR, AVR_X, AVR_X_2016)

DESCRIPTION_TYPES = {
    AVR_NAME: DescriptionType(port=8080, url="/description.xml"),
    AVR_X_NAME: DescriptionType(port=8080, url="/description.xml"),
    AVR_X_2016_NAME: DescriptionType(
        port=60006, url="/upnp/desc/aios_device/aios_device.xml"
    ),
}

SOURCE_MAPPING = {
    "TV AUDIO": "TV",
    "iPod/USB": "USB/IPOD",
    "Bluetooth": "BT",
    "Blu-ray": "BD",
    "CBL/SAT": "SAT/CBL",
    "NETWORK": "NET",
    "Media Player": "MPLAY",
    "AUX": "AUX1",
    "Tuner": "TUNER",
    "FM": "TUNER",
    "SpotifyConnect": "Spotify Connect",
}

CHANGE_INPUT_MAPPING = {
    "Favorites": "FAVORITES",
    "Flickr": "FLICKR",
    "Internet Radio": "IRADIO",
    "Media Server": "SERVER",
    "Online Music": "NET",
    "Spotify": "SPOTIFY",
}

TELNET_SOURCES = {
    "CD",
    "PHONO",
    "TUNER",
    "DVD",
    "BD",
    "TV",
    "SAT/CBL",
    "MPLAY",
    "GAME",
    "HDRADIO",
    "NET",
    "PANDORA",
    "SIRIUSXM",
    "SOURCE",
    "LASTFM",
    "FLICKR",
    "IRADIO",
    "IRP",
    "SERVER",
    "FAVORITES",
    "AUX1",
    "AUX2",
    "AUX3",
    "AUX4",
    "AUX5",
    "AUX6",
    "AUX7",
    "BT",
    "USB/IPOD",
    "USB DIRECT",
    "IPOD DIRECT",
}

TELNET_MAPPING = {
    "FAVORITES": "Favorites",
    "FLICKR": "Flickr",
    "IRADIO": "Internet Radio",
    "IRP": "Internet Radio",
    "SERVER": "Media Server",
    "SPOTIFY": "Spotify",
}

# Sound modes
ALL_ZONE_STEREO = "ALL ZONE STEREO"

SOUND_MODE_MAPPING = {
    "MUSIC": [
        "DOLBY D +NEO:X M",
        "DOLBY PL2 M",
        "DOLBY PL2 MUSIC",
        "DOLBY PL2 X MUSIC",
        "DTS NEO:6 M",
        "DTS NEO:6 MUSIC",
        "DTS NEO:X M",
        "DTS NEO:X MUSIC",
        "PLII MUSIC",
        "PLIIX MUSIC",
    ],
    "MOVIE": [
        "DOLBY D +NEO:X C",
        "DOLBY PL2 C",
        "DOLBY PL2 CINEMA",
        "DOLBY PL2 MOVIE",
        "DOLBY PL2 X MOVIE",
        "DOLBY PLII MOVIE",
        "DTS NEO:6 C",
        "DTS NEO:6 CINEMA",
        "DTS NEO:X C",
        "DTS NEO:X CINEMA",
        "MULTI IN + VIRTUAL:X",
        "PLII CINEMA",
        "PLII MOVIE",
        "PLIIX CINEMA",
    ],
    "GAME": [
        "DOLBY D +NEO:X G",
        "DOLBY PL2 G",
        "DOLBY PL2 GAME",
        "DOLBY PL2 X GAME",
        "DOLBY PLII GAME",
        "PLII GAME",
    ],
    "AUTO": ["None"],
    "STANDARD": ["None2"],
    "VIRTUAL": ["VIRTUAL"],
    "MATRIX": ["MATRIX"],
    "ROCK ARENA": ["ROCK ARENA"],
    "JAZZ CLUB": ["JAZZ CLUB"],
    "VIDEO GAME": ["VIDEO GAME"],
    "MONO MOVIE": ["MONO MOVIE"],
    "DIRECT": ["DIRECT"],
    "PURE DIRECT": ["PURE_DIRECT", "PURE DIRECT"],
    "DOLBY DIGITAL": [
        "DOLBY ATMOS",
        "DOLBY AUDIO - DD + DSUR",
        "DOLBY AUDIO - DD + NEURAL:X",
        "DOLBY AUDIO - DD+   + DSUR",
        "DOLBY AUDIO - DD+   + NEURAL:X",
        "DOLBY AUDIO - DOLBY DIGITAL PLUS",
        "DOLBY AUDIO - DOLBY DIGITAL",
        "DOLBY AUDIO - DOLBY SURROUND",
        "DOLBY AUDIO - DOLBY TRUEHD",
        "DOLBY AUDIO - TRUEHD + DSUR",
        "DOLBY AUDIO - TRUEHD + NEURAL:X",
        "DOLBY AUDIO-DD",
        "DOLBY AUDIO-DD+ +DSUR",
        # Added both variants as we don't know if type is intentional
        "DOLBY AUDIO-DD+ +NERUAL:X",
        "DOLBY AUDIO-DD+ +NEURAL:X",
        "DOLBY AUDIO-DD+",
        "DOLBY AUDIO-DD+",
        "DOLBY AUDIO-DD+DSUR",
        "DOLBY AUDIO-DD+NEURAL:X",
        "DOLBY AUDIO-DSUR",
        "DOLBY AUDIO-TRUEHD",
        "DOLBY AUDIO-TRUEHD+DSUR",
        "DOLBY AUDIO-TRUEHD+NEURAL:X",
        "DOLBY D + +DOLBY SURROUND",
        "DOLBY D + DOLBY SURROUND",
        "DOLBY D + NEURAL:X",
        "DOLBY D+ +DS",
        "DOLBY D+",
        "DOLBY D+DS",
        "DOLBY DIGITAL + + NEURAL:X",
        "DOLBY DIGITAL + NEURAL:X",
        "DOLBY DIGITAL +",
        "DOLBY DIGITAL",
        "DOLBY HD + DOLBY SURROUND",
        "DOLBY HD",
        "DOLBY PRO LOGIC",
        "DOLBY SURROUND",
        "DOLBY TRUEHD",
        "MULTI IN + DOLBY SURROUND",
        "MULTI IN + DSUR",
        "MULTI IN + NEURAL:X",
        "NEURAL",
        "STANDARD(DOLBY)",
    ],
    "DTS SURROUND": [
        "DTS + DOLBY SURROUND",
        "DTS + NEURAL:X",
        "DTS + VIRTUAL:X",
        "DTS HD MSTR",
        "DTS HD",
        "DTS HD+DSUR",
        "DTS HD+NEURAL:X",
        "DTS HD+VIRTUAL:X",
        "DTS NEURAL:X",
        "DTS SURROUND",
        "DTS VIRTUAL:X",
        "DTS-HD + DOLBY SURROUND",
        "DTS-HD + DSUR",
        "DTS-HD + NEURAL:X",
        "DTS-HD MSTR",
        "DTS-HD",
        "DTS:X MSTR",
        "DTS:X",
        "DTS:X+VIRTUAL:X",
        "DTS+DSUR",
        "DTS+NEURAL:X",
        "DTS+VIRTUAL:X",
        "IMAX DTS:X",
        "IMAX DTS:X+VIRTUAL:X",
        "IMAX DTS",
        "IMAX DTS+NEURAL:X",
        "M CH IN+DSUR",
        "M CH IN+NEURAL:X",
        "M CH IN+VIRTUAL:X",
        "MAX DTS+VIRTUAL:X",
        "MULTI CH IN",
        "NEURAL:X",
        "STANDARD(DTS)",
        "VIRTUAL:X",
    ],
    "AURO3D": ["AURO-3D", "AURO3D"],
    "AURO2DSURR": ["AURO-2D SURROUND", "AURO2DSURR"],
    "MCH STEREO": [
        "MCH STEREO",
        "MULTI CH IN 7.1",
        "MULTI CH STEREO",
        "MULTI_CH_STEREO",
    ],
    "STEREO": ["STEREO"],
    ALL_ZONE_STEREO: ["ALL ZONE STEREO"],
}

# Receiver sources
NETAUDIO_SOURCES = {
    "AirPlay",
    "Online Music",
    "Media Server",
    "iPod/USB",
    "Bluetooth",
    "Internet Radio",
    "Favorites",
    "SpotifyConnect",
    "Flickr",
    "NET/USB",
    "Music Server",
    "NETWORK",
    "NET",
}
TUNER_SOURCES = {"Tuner", "TUNER"}
HDTUNER_SOURCES = {"HD Radio", "HDRADIO"}
PLAYING_SOURCES = set().union(*[NETAUDIO_SOURCES, TUNER_SOURCES, HDTUNER_SOURCES])

# Image URLs
STATIC_ALBUM_URL = "http://{host}:{port}/img/album%20art_S.png"
ALBUM_COVERS_URL = "http://{host}:{port}/NetAudio/art.asp-jpg?{hash}"

# General URLs
APPCOMMAND_URL = "/goform/AppCommand.xml"
APPCOMMAND0300_URL = "/goform/AppCommand0300.xml"
DEVICEINFO_URL = "/goform/Deviceinfo.xml"
NETAUDIOSTATUS_URL = "/goform/formNetAudio_StatusXml.xml"
TUNERSTATUS_URL = "/goform/formTuner_TunerXml.xml"
HDTUNERSTATUS_URL = "/goform/formTuner_HdXml.xml"
COMMAND_NETAUDIO_POST_URL = "/NetAudio/index.put.asp"
COMMAND_PAUSE = "/goform/formiPhoneAppDirect.xml?NS9B"
COMMAND_PLAY = "/goform/formiPhoneAppDirect.xml?NS9A"
COMMAND_ECO_MODE = "/goform/formiPhoneAppDirect.xml?ECO{mode}"


# Main Zone URLs
STATUS_URL = "/goform/formMainZone_MainZoneXmlStatus.xml"
MAINZONE_URL = "/goform/formMainZone_MainZoneXml.xml"
COMMAND_SEL_SRC_URL = "/goform/formiPhoneAppDirect.xml?SI"
COMMAND_FAV_SRC_URL = "/goform/formiPhoneAppDirect.xml?ZM"
COMMAND_POWER_ON_URL = "/goform/formiPhoneAppPower.xml?1+PowerOn"
COMMAND_POWER_STANDBY_URL = "/goform/formiPhoneAppPower.xml?1+PowerStandby"
COMMAND_VOLUME_UP_URL = "/goform/formiPhoneAppDirect.xml?MVUP"
COMMAND_VOLUME_DOWN_URL = "/goform/formiPhoneAppDirect.xml?MVDOWN"
COMMAND_SET_VOLUME_URL = "/goform/formiPhoneAppVolume.xml?1+{volume:.1f}"
COMMAND_MUTE_ON_URL = "/goform/formiPhoneAppMute.xml?1+MuteOn"
COMMAND_MUTE_OFF_URL = "/goform/formiPhoneAppMute.xml?1+MuteOff"
COMMAND_SEL_SM_URL = "/goform/formiPhoneAppDirect.xml?MS"
COMMAND_SET_ZST_URL = "/goform/formiPhoneAppDirect.xml?MN"
COMMAND_NEURAL_X_ON_OFF = "/goform/formiPhoneAppDirect.xml?PSNEURAL%20{mode}"
COMMAND_IMAX_AUTO_OFF = "/goform/formiPhoneAppDirect.xml?PSIMAX%20{mode}"
COMMAND_IMAX_AUDIO_SETTINGS = "/goform/formiPhoneAppDirect.xml?PSIMAXAUD%20{mode}"
COMMAND_IMAX_HPF = "/goform/formiPhoneAppDirect.xml?PSIMAXHPF%20{frequency}"
COMMAND_IMAX_LPF = "/goform/formiPhoneAppDirect.xml?PSIMAXLPF%20{frequency}"
COMMAND_IMAX_SUBWOOFER_MODE = "/goform/formiPhoneAppDirect.xml?PSIMAXSWM%20{mode}"
COMMAND_IMAX_SUBWOOFER_OUTPUT = "/goform/formiPhoneAppDirect.xml?PSIMAXSWO%20{mode}"
COMMAND_CINEMA_EQ = "/goform/formiPhoneAppDirect.xml?PSCINEMA%20EQ.{mode}"
COMMAND_CURSOR_UP = "/goform/formiPhoneAppDirect.xml?MNCUP"
COMMAND_CURSOR_DOWN = "/goform/formiPhoneAppDirect.xml?MNCDN"
COMMAND_CURSOR_LEFT = "/goform/formiPhoneAppDirect.xml?MNCLT"
COMMAND_CURSOR_RIGHT = "/goform/formiPhoneAppDirect.xml?MNCRT"
COMMAND_CURSOR_ENTER = "/goform/formiPhoneAppDirect.xml?MNENT"
COMMAND_BACK = "/goform/formiPhoneAppDirect.xml?MNRTN"
COMMAND_INFO = "/goform/formiPhoneAppDirect.xml?MNINF"
COMMAND_OPTIONS = "/goform/formiPhoneAppDirect.xml?MNOPT"
COMMAND_SETUP_OPEN = "/goform/formiPhoneAppDirect.xml?MNMEN%20ON"
COMMAND_SETUP_CLOSE = "/goform/formiPhoneAppDirect.xml?MNMEN%20OFF"
COMMAND_SETUP_QUERY = "/goform/formiPhoneAppDirect.xml?MNMEN?"
COMMAND_CHANNEL_LEVEL_ADJUST = "/goform/formiPhoneAppDirect.xml?MNCHL"
COMMAND_DIMMER_TOGGLE = "/goform/formiPhoneAppDirect.xml?DIM%20SEL"
COMMAND_DIMMER_SET = "/goform/formiPhoneAppDirect.xml?DIM%20{mode}"
COMMAND_CHANNEL_VOLUME = "/goform/formiPhoneAppDirect.xml?CV{channel}%20{value}"
COMMAND_CHANNEL_VOLUMES_RESET = "/goform/formiPhoneAppDirect.xml?CVZRL"
COMMAND_SUBWOOFER_ON_OFF = "/goform/formiPhoneAppDirect.xml?PSSWR%20{mode}"
COMMAND_SUBWOOFER_LEVEL = "/goform/formiPhoneAppDirect.xml?PSSWL{number}%20{mode}"
COMMAND_LFE = "/goform/formiPhoneAppDirect.xml?PSLFE%20{mode}"
COMMAND_TACTILE_TRANSDUCER = "/goform/formiPhoneAppDirect.xml?SSTTR%20{mode}"
COMMAND_TACTILE_TRANSDUCER_LEVEL = "/goform/formiPhoneAppDirect.xml?SSTTRLEV%20{mode}"
COMMAND_TACTILE_TRANSDUCER_LPF = (
    "/goform/formiPhoneAppDirect.xml?SSTTRLPF%20{frequency}"
)
COMMAND_DELAY_UP = "/goform/formiPhoneAppDirect.xml?PSDELAY%20UP"
COMMAND_DELAY_DOWN = "/goform/formiPhoneAppDirect.xml?PSDELAY%20DOWN"
COMMAND_AUROMATIC_3D_PRESET = "/goform/formiPhoneAppDirect.xml?PSAUROPR%20{preset}"
COMMAND_AUROMATIC_3D_STRENGTH = "/goform/formiPhoneAppDirect.xml?PSAUROST%20{value}"
COMMAND_AURO_3D_MODE = "/goform/formiPhoneAppDirect.xml?PSAUROMODEM%20{mode}"
COMMAND_DIRAC_FILTER = "/goform/formiPhoneAppDirect.xml?PSDIRAC%20{filter}"
COMMAND_LFC = "/goform/formiPhoneAppDirect.xml?PSLFC%20{mode}"
COMMAND_CONTAMINATION_AMOUNT = "/goform/formiPhoneAppDirect.xml?PSCNTAMT%20{value}"
COMMAND_LOUDNESS_MANAGEMENT = "/goform/formiPhoneAppDirect.xml?PSLOM%20{mode}"
COMMAND_BASS_SYNC = "/goform/formiPhoneAppDirect.xml?PSBSC%20{mode}"
COMMAND_DIALOG_ENHANCER = "/goform/formiPhoneAppDirect.xml?PSDEH%20{level}"
COMMAND_HDMI_OUTPUT = "/goform/formiPhoneAppDirect.xml?VSMONI{output}"
COMMAND_HDMI_AUDIO_DECODE = "/goform/formiPhoneAppDirect.xml?VSAUDIO%20{mode}"
COMMAND_QUICK_SELECT_MODE = "/goform/formiPhoneAppDirect.xml?MSQUICK{number}"
COMMAND_QUICK_SELECT_MEMORY = "/goform/formiPhoneAppDirect.xml?MSQUICK{number}%20MEMORY"
COMMAND_SMART_SELECT_MODE = "/goform/formiPhoneAppDirect.xml?MSSMART{number}"  # Marantz
COMMAND_SMART_SELECT_MEMORY = (
    "/goform/formiPhoneAppDirect.xml?MSSMART{number}%20MEMORY"  # Marantz
)
COMMAND_AUTO_STANDBY = "/goform/formiPhoneAppDirect.xml?STBY{mode}"
COMMAND_SLEEP = "/goform/formiPhoneAppDirect.xml?SLP{value}"
COMMAND_CENTER_SPREAD = "/goform/formiPhoneAppDirect.xml?PSCES%20{mode}"
COMMAND_VIDEO_PROCESSING_MODE = "/goform/formiPhoneAppDirect.xml?VSVPM{mode}"
COMMAND_ROOM_SIZE = "/goform/formiPhoneAppDirect.xml?PSRSZ%20{size}"
COMMAND_STATUS = "/goform/formiPhoneAppDirect.xml?RCSHP0230030"
COMMAND_SYSTEM_RESET = "/goform/formiPhoneAppDirect.xml?SYRST"
COMMAND_NETWORK_RESTART = "/goform/formiPhoneAppDirect.xml?NSRBT"
COMMAND_TRIGGER = "/goform/formiPhoneAppDirect.xml?TR{number}%20{mode}"
COMMAND_SPEAKER_PRESET = "/goform/formiPhoneAppDirect.xml?SPPR%20{number}"
COMMAND_BLUETOOTH_TRANSMITTER = "/goform/formiPhoneAppDirect.xml?BTTX%20{mode}"
COMMAND_DIALOG_CONTROL = "/goform/formiPhoneAppDirect.xml?PSDIC%20{value}"
COMMAND_SPEAKER_VIRTUALIZER = "/goform/formiPhoneAppDirect.xml?PSSPV%20{mode}"
COMMAND_EFFECT_SPEAKER_SELECTION = "/goform/formiPhoneAppDirect.xml?PSSP:{mode}"
COMMAND_DRC = "/goform/formiPhoneAppDirect.xml?PSDRC%20{mode}"
COMMAND_DELAY_TIME = "/goform/formiPhoneAppDirect.xml?PSDEL%20{value}"
COMMAND_AUDIO_RESTORER = "/goform/formiPhoneAppDirect.xml?PSRSTR%20{mode}"
COMMAND_REMOTE_CONTROL_LOCK = "/goform/formiPhoneAppDirect.xml?SYREMOTE%20LOCK%20{mode}"
COMMAND_PANEL_LOCK = "/goform/formiPhoneAppDirect.xml?SYPANEL%20LOCK%20{mode}"
COMMAND_PANEL_AND_VOLUME_LOCK = "/goform/formiPhoneAppDirect.xml?SYPANEL+V%20LOCK%20ON"
COMMAND_GRAPHIC_EQ = "/goform/formiPhoneAppDirect.xml?PSGEQ%20{mode}"
COMMAND_HEADPHONE_EQ = "/goform/formiPhoneAppDirect.xml?PSHEQ%20{mode}"

# Zone 2 URLs
STATUS_Z2_URL = "/goform/formZone2_Zone2XmlStatus.xml"
COMMAND_SEL_SRC_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2"
COMMAND_FAV_SRC_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2"
COMMAND_POWER_ON_Z2_URL = "/goform/formiPhoneAppPower.xml?2+PowerOn"
COMMAND_POWER_STANDBY_Z2_URL = "/goform/formiPhoneAppPower.xml?2+PowerStandby"
COMMAND_VOLUME_UP_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2UP"
COMMAND_VOLUME_DOWN_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2DOWN"
COMMAND_SET_VOLUME_Z2_URL = "/goform/formiPhoneAppVolume.xml?2+{volume:.1f}"
COMMAND_MUTE_ON_Z2_URL = "/goform/formiPhoneAppMute.xml?2+MuteOn"
COMMAND_MUTE_OFF_Z2_URL = "/goform/formiPhoneAppMute.xml?2+MuteOff"

# Zone 3 URLs
STATUS_Z3_URL = "/goform/formZone3_Zone3XmlStatus.xml"
COMMAND_SEL_SRC_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3"
COMMAND_FAV_SRC_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3"
COMMAND_POWER_ON_Z3_URL = "/goform/formiPhoneAppPower.xml?3+PowerOn"
COMMAND_POWER_STANDBY_Z3_URL = "/goform/formiPhoneAppPower.xml?3+PowerStandby"
COMMAND_VOLUME_UP_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3UP"
COMMAND_VOLUME_DOWN_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3DOWN"
COMMAND_SET_VOLUME_Z3_URL = "/goform/formiPhoneAppVolume.xml?3+{volume:.1f}"
COMMAND_MUTE_ON_Z3_URL = "/goform/formiPhoneAppMute.xml?3+MuteOn"
COMMAND_MUTE_OFF_Z3_URL = "/goform/formiPhoneAppMute.xml?3+MuteOff"

DENONAVR_URLS = ReceiverURLs(
    appcommand=APPCOMMAND_URL,
    appcommand0300=APPCOMMAND0300_URL,
    status=STATUS_URL,
    mainzone=MAINZONE_URL,
    deviceinfo=DEVICEINFO_URL,
    netaudiostatus=NETAUDIOSTATUS_URL,
    tunerstatus=TUNERSTATUS_URL,
    hdtunerstatus=HDTUNERSTATUS_URL,
    command_sel_src=COMMAND_SEL_SRC_URL,
    command_fav_src=COMMAND_FAV_SRC_URL,
    command_power_on=COMMAND_POWER_ON_URL,
    command_power_standby=COMMAND_POWER_STANDBY_URL,
    command_volume_up=COMMAND_VOLUME_UP_URL,
    command_volume_down=COMMAND_VOLUME_DOWN_URL,
    command_set_volume=COMMAND_SET_VOLUME_URL,
    command_mute_on=COMMAND_MUTE_ON_URL,
    command_mute_off=COMMAND_MUTE_OFF_URL,
    command_sel_sound_mode=COMMAND_SEL_SM_URL,
    command_neural_x_on_off=COMMAND_NEURAL_X_ON_OFF,
    command_imax_auto_off=COMMAND_IMAX_AUTO_OFF,
    command_imax_audio_settings=COMMAND_IMAX_AUTO_OFF,
    command_imax_hpf=COMMAND_IMAX_HPF,
    command_imax_lpf=COMMAND_IMAX_LPF,
    command_imax_subwoofer_mode=COMMAND_IMAX_SUBWOOFER_MODE,
    command_imax_subwoofer_output=COMMAND_IMAX_SUBWOOFER_OUTPUT,
    command_cinema_eq=COMMAND_CINEMA_EQ,
    command_netaudio_post=COMMAND_NETAUDIO_POST_URL,
    command_set_all_zone_stereo=COMMAND_SET_ZST_URL,
    command_pause=COMMAND_PAUSE,
    command_play=COMMAND_PLAY,
    command_cusor_up=COMMAND_CURSOR_UP,
    command_cusor_down=COMMAND_CURSOR_DOWN,
    command_cusor_left=COMMAND_CURSOR_LEFT,
    command_cusor_right=COMMAND_CURSOR_RIGHT,
    command_cusor_enter=COMMAND_CURSOR_ENTER,
    command_back=COMMAND_BACK,
    command_info=COMMAND_INFO,
    command_options=COMMAND_OPTIONS,
    command_setup_open=COMMAND_SETUP_OPEN,
    command_setup_close=COMMAND_SETUP_CLOSE,
    command_setup_query=COMMAND_SETUP_QUERY,
    command_channel_level_adjust=COMMAND_CHANNEL_LEVEL_ADJUST,
    command_dimmer_toggle=COMMAND_DIMMER_TOGGLE,
    command_dimmer_set=COMMAND_DIMMER_SET,
    command_channel_volume=COMMAND_CHANNEL_VOLUME,
    command_channel_volumes_reset=COMMAND_CHANNEL_VOLUMES_RESET,
    command_subwoofer_on_off=COMMAND_SUBWOOFER_ON_OFF,
    command_subwoofer_level=COMMAND_SUBWOOFER_LEVEL,
    command_lfe=COMMAND_LFE,
    command_tactile_transducer=COMMAND_TACTILE_TRANSDUCER,
    command_tactile_transducer_level=COMMAND_TACTILE_TRANSDUCER_LEVEL,
    command_tactile_transducer_lpf=COMMAND_TACTILE_TRANSDUCER_LPF,
    command_delay_up=COMMAND_DELAY_UP,
    command_delay_down=COMMAND_DELAY_DOWN,
    command_auromatic_3d_preset=COMMAND_AUROMATIC_3D_PRESET,
    command_auromatic_3d_strength=COMMAND_AUROMATIC_3D_STRENGTH,
    command_auro_3d_mode=COMMAND_AURO_3D_MODE,
    command_dirac_filter=COMMAND_DIRAC_FILTER,
    command_eco_mode=COMMAND_ECO_MODE,
    command_lfc=COMMAND_LFC,
    command_containment_amount=COMMAND_CONTAMINATION_AMOUNT,
    command_loudness_management=COMMAND_LOUDNESS_MANAGEMENT,
    command_bass_sync=COMMAND_BASS_SYNC,
    command_dialog_enhancer=COMMAND_DIALOG_ENHANCER,
    command_hdmi_output=COMMAND_HDMI_OUTPUT,
    command_hdmi_audio_decode=COMMAND_HDMI_AUDIO_DECODE,
    command_quick_select_mode=COMMAND_QUICK_SELECT_MODE,
    command_quick_select_memory=COMMAND_QUICK_SELECT_MODE,
    command_smart_select_mode=COMMAND_SMART_SELECT_MODE,  # Marantz
    command_smart_select_memory=COMMAND_SMART_SELECT_MEMORY,  # Marantz
    command_auto_standby=COMMAND_AUTO_STANDBY,
    command_sleep=COMMAND_SLEEP,
    command_center_spread=COMMAND_CENTER_SPREAD,
    command_video_processing_mode=COMMAND_VIDEO_PROCESSING_MODE,
    command_room_size=COMMAND_ROOM_SIZE,
    command_status=COMMAND_STATUS,
    command_system_reset=COMMAND_SYSTEM_RESET,
    command_network_restart=COMMAND_NETWORK_RESTART,
    command_trigger=COMMAND_TRIGGER,
    command_speaker_preset=COMMAND_SPEAKER_PRESET,
    command_bluetooth_transmitter=COMMAND_BLUETOOTH_TRANSMITTER,
    command_dialog_control=COMMAND_DIALOG_CONTROL,
    command_speaker_virtualizer=COMMAND_SPEAKER_VIRTUALIZER,
    command_effect_speaker_selection=COMMAND_EFFECT_SPEAKER_SELECTION,
    command_drc=COMMAND_DRC,
    command_delay_time=COMMAND_DELAY_TIME,
    command_audio_restorer=COMMAND_AUDIO_RESTORER,
    command_remote_control_lock=COMMAND_REMOTE_CONTROL_LOCK,
    command_panel_lock=COMMAND_PANEL_LOCK,
    command_panel_and_volume_lock=COMMAND_PANEL_AND_VOLUME_LOCK,
    command_graphic_eq=COMMAND_GRAPHIC_EQ,
    command_headphone_eq=COMMAND_HEADPHONE_EQ,
)

ZONE2_URLS = ReceiverURLs(
    appcommand=APPCOMMAND_URL,
    appcommand0300=APPCOMMAND0300_URL,
    status=STATUS_Z2_URL,
    mainzone=MAINZONE_URL,
    deviceinfo=DEVICEINFO_URL,
    netaudiostatus=NETAUDIOSTATUS_URL,
    tunerstatus=TUNERSTATUS_URL,
    hdtunerstatus=HDTUNERSTATUS_URL,
    command_sel_src=COMMAND_SEL_SRC_Z2_URL,
    command_fav_src=COMMAND_FAV_SRC_Z2_URL,
    command_power_on=COMMAND_POWER_ON_Z2_URL,
    command_power_standby=COMMAND_POWER_STANDBY_Z2_URL,
    command_volume_up=COMMAND_VOLUME_UP_Z2_URL,
    command_volume_down=COMMAND_VOLUME_DOWN_Z2_URL,
    command_set_volume=COMMAND_SET_VOLUME_Z2_URL,
    command_mute_on=COMMAND_MUTE_ON_Z2_URL,
    command_mute_off=COMMAND_MUTE_OFF_Z2_URL,
    command_sel_sound_mode=COMMAND_SEL_SM_URL,
    command_neural_x_on_off=COMMAND_NEURAL_X_ON_OFF,
    command_imax_auto_off=COMMAND_IMAX_AUTO_OFF,
    command_imax_audio_settings=COMMAND_IMAX_AUDIO_SETTINGS,
    command_imax_hpf=COMMAND_IMAX_HPF,
    command_imax_lpf=COMMAND_IMAX_LPF,
    command_imax_subwoofer_mode=COMMAND_IMAX_SUBWOOFER_MODE,
    command_imax_subwoofer_output=COMMAND_IMAX_SUBWOOFER_OUTPUT,
    command_cinema_eq=COMMAND_CINEMA_EQ,
    command_netaudio_post=COMMAND_NETAUDIO_POST_URL,
    command_set_all_zone_stereo=COMMAND_SET_ZST_URL,
    command_pause=COMMAND_PAUSE,
    command_play=COMMAND_PLAY,
    command_cusor_up=COMMAND_CURSOR_UP,
    command_cusor_down=COMMAND_CURSOR_DOWN,
    command_cusor_left=COMMAND_CURSOR_LEFT,
    command_cusor_right=COMMAND_CURSOR_RIGHT,
    command_cusor_enter=COMMAND_CURSOR_ENTER,
    command_back=COMMAND_BACK,
    command_info=COMMAND_INFO,
    command_options=COMMAND_OPTIONS,
    command_setup_open=COMMAND_SETUP_OPEN,
    command_setup_close=COMMAND_SETUP_CLOSE,
    command_setup_query=COMMAND_SETUP_QUERY,
    command_channel_level_adjust=COMMAND_CHANNEL_LEVEL_ADJUST,
    command_dimmer_toggle=COMMAND_DIMMER_TOGGLE,
    command_dimmer_set=COMMAND_DIMMER_SET,
    command_channel_volume=COMMAND_CHANNEL_VOLUME,
    command_channel_volumes_reset=COMMAND_CHANNEL_VOLUMES_RESET,
    command_subwoofer_on_off=COMMAND_SUBWOOFER_ON_OFF,
    command_subwoofer_level=COMMAND_SUBWOOFER_LEVEL,
    command_lfe=COMMAND_LFE,
    command_tactile_transducer=COMMAND_TACTILE_TRANSDUCER,
    command_tactile_transducer_level=COMMAND_TACTILE_TRANSDUCER_LEVEL,
    command_tactile_transducer_lpf=COMMAND_TACTILE_TRANSDUCER_LPF,
    command_delay_up=COMMAND_DELAY_UP,
    command_delay_down=COMMAND_DELAY_DOWN,
    command_auromatic_3d_preset=COMMAND_AUROMATIC_3D_PRESET,
    command_auromatic_3d_strength=COMMAND_AUROMATIC_3D_STRENGTH,
    command_auro_3d_mode=COMMAND_AURO_3D_MODE,
    command_dirac_filter=COMMAND_DIRAC_FILTER,
    command_eco_mode=COMMAND_ECO_MODE,
    command_lfc=COMMAND_LFC,
    command_containment_amount=COMMAND_CONTAMINATION_AMOUNT,
    command_loudness_management=COMMAND_LOUDNESS_MANAGEMENT,
    command_bass_sync=COMMAND_BASS_SYNC,
    command_dialog_enhancer=COMMAND_DIALOG_ENHANCER,
    command_hdmi_output=COMMAND_HDMI_OUTPUT,
    command_hdmi_audio_decode=COMMAND_HDMI_AUDIO_DECODE,
    command_quick_select_mode=COMMAND_QUICK_SELECT_MODE,
    command_quick_select_memory=COMMAND_QUICK_SELECT_MEMORY,
    command_smart_select_mode=COMMAND_SMART_SELECT_MODE,  # Marantz
    command_smart_select_memory=COMMAND_SMART_SELECT_MEMORY,  # Marantz
    command_auto_standby=COMMAND_AUTO_STANDBY,
    command_sleep=COMMAND_SLEEP,
    command_center_spread=COMMAND_CENTER_SPREAD,
    command_video_processing_mode=COMMAND_VIDEO_PROCESSING_MODE,
    command_room_size=COMMAND_ROOM_SIZE,
    command_status=COMMAND_STATUS,
    command_system_reset=COMMAND_SYSTEM_RESET,
    command_network_restart=COMMAND_NETWORK_RESTART,
    command_trigger=COMMAND_TRIGGER,
    command_speaker_preset=COMMAND_SPEAKER_PRESET,
    command_bluetooth_transmitter=COMMAND_BLUETOOTH_TRANSMITTER,
    command_dialog_control=COMMAND_DIALOG_CONTROL,
    command_speaker_virtualizer=COMMAND_SPEAKER_VIRTUALIZER,
    command_effect_speaker_selection=COMMAND_EFFECT_SPEAKER_SELECTION,
    command_drc=COMMAND_DRC,
    command_delay_time=COMMAND_DELAY_TIME,
    command_audio_restorer=COMMAND_AUDIO_RESTORER,
    command_remote_control_lock=COMMAND_REMOTE_CONTROL_LOCK,
    command_panel_lock=COMMAND_PANEL_LOCK,
    command_panel_and_volume_lock=COMMAND_PANEL_AND_VOLUME_LOCK,
    command_graphic_eq=COMMAND_GRAPHIC_EQ,
    command_headphone_eq=COMMAND_HEADPHONE_EQ,
)

ZONE3_URLS = ReceiverURLs(
    appcommand=APPCOMMAND_URL,
    appcommand0300=APPCOMMAND0300_URL,
    status=STATUS_Z3_URL,
    mainzone=MAINZONE_URL,
    deviceinfo=DEVICEINFO_URL,
    netaudiostatus=NETAUDIOSTATUS_URL,
    tunerstatus=TUNERSTATUS_URL,
    hdtunerstatus=HDTUNERSTATUS_URL,
    command_sel_src=COMMAND_SEL_SRC_Z3_URL,
    command_fav_src=COMMAND_FAV_SRC_Z3_URL,
    command_power_on=COMMAND_POWER_ON_Z3_URL,
    command_power_standby=COMMAND_POWER_STANDBY_Z3_URL,
    command_volume_up=COMMAND_VOLUME_UP_Z3_URL,
    command_volume_down=COMMAND_VOLUME_DOWN_Z3_URL,
    command_set_volume=COMMAND_SET_VOLUME_Z3_URL,
    command_mute_on=COMMAND_MUTE_ON_Z3_URL,
    command_mute_off=COMMAND_MUTE_OFF_Z3_URL,
    command_sel_sound_mode=COMMAND_SEL_SM_URL,
    command_neural_x_on_off=COMMAND_NEURAL_X_ON_OFF,
    command_imax_auto_off=COMMAND_IMAX_AUTO_OFF,
    command_imax_audio_settings=COMMAND_IMAX_AUDIO_SETTINGS,
    command_imax_hpf=COMMAND_IMAX_HPF,
    command_imax_lpf=COMMAND_IMAX_LPF,
    command_imax_subwoofer_mode=COMMAND_IMAX_SUBWOOFER_MODE,
    command_imax_subwoofer_output=COMMAND_IMAX_SUBWOOFER_OUTPUT,
    command_cinema_eq=COMMAND_CINEMA_EQ,
    command_netaudio_post=COMMAND_NETAUDIO_POST_URL,
    command_set_all_zone_stereo=COMMAND_SET_ZST_URL,
    command_pause=COMMAND_PAUSE,
    command_play=COMMAND_PLAY,
    command_cusor_up=COMMAND_CURSOR_UP,
    command_cusor_down=COMMAND_CURSOR_DOWN,
    command_cusor_left=COMMAND_CURSOR_LEFT,
    command_cusor_right=COMMAND_CURSOR_RIGHT,
    command_cusor_enter=COMMAND_CURSOR_ENTER,
    command_back=COMMAND_BACK,
    command_info=COMMAND_INFO,
    command_options=COMMAND_OPTIONS,
    command_setup_open=COMMAND_SETUP_OPEN,
    command_setup_close=COMMAND_SETUP_CLOSE,
    command_setup_query=COMMAND_SETUP_QUERY,
    command_channel_level_adjust=COMMAND_CHANNEL_LEVEL_ADJUST,
    command_dimmer_toggle=COMMAND_DIMMER_TOGGLE,
    command_dimmer_set=COMMAND_DIMMER_SET,
    command_channel_volume=COMMAND_CHANNEL_VOLUME,
    command_channel_volumes_reset=COMMAND_CHANNEL_VOLUMES_RESET,
    command_subwoofer_on_off=COMMAND_SUBWOOFER_ON_OFF,
    command_subwoofer_level=COMMAND_SUBWOOFER_LEVEL,
    command_lfe=COMMAND_LFE,
    command_tactile_transducer=COMMAND_TACTILE_TRANSDUCER,
    command_tactile_transducer_level=COMMAND_TACTILE_TRANSDUCER_LEVEL,
    command_tactile_transducer_lpf=COMMAND_TACTILE_TRANSDUCER_LPF,
    command_delay_up=COMMAND_DELAY_UP,
    command_delay_down=COMMAND_DELAY_DOWN,
    command_auromatic_3d_preset=COMMAND_AUROMATIC_3D_PRESET,
    command_auromatic_3d_strength=COMMAND_AUROMATIC_3D_STRENGTH,
    command_auro_3d_mode=COMMAND_AURO_3D_MODE,
    command_dirac_filter=COMMAND_DIRAC_FILTER,
    command_eco_mode=COMMAND_ECO_MODE,
    command_lfc=COMMAND_LFC,
    command_containment_amount=COMMAND_CONTAMINATION_AMOUNT,
    command_loudness_management=COMMAND_LOUDNESS_MANAGEMENT,
    command_bass_sync=COMMAND_BASS_SYNC,
    command_dialog_enhancer=COMMAND_DIALOG_ENHANCER,
    command_hdmi_output=COMMAND_HDMI_OUTPUT,
    command_hdmi_audio_decode=COMMAND_HDMI_AUDIO_DECODE,
    command_quick_select_mode=COMMAND_QUICK_SELECT_MODE,
    command_quick_select_memory=COMMAND_QUICK_SELECT_MEMORY,
    command_smart_select_mode=COMMAND_SMART_SELECT_MODE,  # Marantz
    command_smart_select_memory=COMMAND_SMART_SELECT_MEMORY,  # Marantz
    command_auto_standby=COMMAND_AUTO_STANDBY,
    command_sleep=COMMAND_SLEEP,
    command_center_spread=COMMAND_CENTER_SPREAD,
    command_video_processing_mode=COMMAND_VIDEO_PROCESSING_MODE,
    command_room_size=COMMAND_ROOM_SIZE,
    command_status=COMMAND_STATUS,
    command_system_reset=COMMAND_SYSTEM_RESET,
    command_network_restart=COMMAND_NETWORK_RESTART,
    command_trigger=COMMAND_TRIGGER,
    command_speaker_preset=COMMAND_SPEAKER_PRESET,
    command_bluetooth_transmitter=COMMAND_BLUETOOTH_TRANSMITTER,
    command_dialog_control=COMMAND_DIALOG_CONTROL,
    command_speaker_virtualizer=COMMAND_SPEAKER_VIRTUALIZER,
    command_effect_speaker_selection=COMMAND_EFFECT_SPEAKER_SELECTION,
    command_drc=COMMAND_DRC,
    command_delay_time=COMMAND_DELAY_TIME,
    command_audio_restorer=COMMAND_AUDIO_RESTORER,
    command_remote_control_lock=COMMAND_REMOTE_CONTROL_LOCK,
    command_panel_lock=COMMAND_PANEL_LOCK,
    command_panel_and_volume_lock=COMMAND_PANEL_AND_VOLUME_LOCK,
    command_graphic_eq=COMMAND_GRAPHIC_EQ,
    command_headphone_eq=COMMAND_HEADPHONE_EQ,
)

# Telnet Events
ALL_TELNET_EVENTS = "ALL"
TELNET_EVENTS = {
    "BT",
    "CV",
    "DC",
    "DIM",
    "ECO",
    "HD",
    "MN",
    "MS",
    "MU",
    "MV",
    "NS",
    "NSA",
    "NSE",
    "OP",
    "PS",
    "PV",
    "PW",
    "RM",
    "SD",
    "SI",
    "SLP",
    "SP",
    "SR",
    "SS",
    "STBY",
    "SV",
    "SY",
    "TF",
    "TM",
    "TP",
    "TR",
    "UG",
    "VS",
    "ZM",
    "Z2",
    "Z3",
}
ALL_ZONE_TELNET_EVENTS = {
    "DIM",
    "HD",
    "NS",
    "NSA",
    "NSE",
    "MN",
    "PW",
    "RM",
    "SY",
    "TF",
    "TM",
    "TP",
    "TR",
    "UG",
}

DENONAVR_TELNET_COMMANDS = TelnetCommands(
    command_sel_src="SI",
    command_fav_src="ZM",
    command_power_on="ZMON",
    command_power_standby="ZMOFF",
    command_volume_up="MVUP",
    command_volume_down="MVDOWN",
    command_set_volume="MV{volume:02d}",
    command_mute_on="MUON",
    command_mute_off="MUOFF",
    command_sel_sound_mode="MS",
    command_neural_x_on_off="PSNEURAL {mode}",
    command_imax_auto_off="PSIMAX {mode}",
    command_imax_audio_settings="PSIMAXAUD {mode}",
    command_imax_hpf="PSIMAXHPF {frequency}",
    command_imax_lpf="PSIMAXLPF {frequency}",
    command_imax_subwoofer_mode="PSIMAXSWM {mode}",
    command_imax_subwoofer_output="PSIMAXSWO {mode}",
    command_cinema_eq="PSCINEMA EQ.{mode}",
    command_set_all_zone_stereo="MN",
    command_pause="NS9B",
    command_play="NS9A",
    command_multieq="PSMULTEQ:",
    command_dynamiceq="PSDYNEQ ",
    command_reflevoffset="PSREFLEV ",
    command_dynamicvol="PSDYNVOL ",
    command_tonecontrol="PSTONE CTRL ",
    command_bass="PSBAS ",
    command_treble="PSTRE ",
    command_cusor_up="MNCUP",
    command_cusor_down="MNCDN",
    command_cusor_left="MNCLT",
    command_cusor_right="MNCRT",
    command_cusor_enter="MNENT",
    command_back="MNRTN",
    command_info="MNINF",
    command_options="MNOPT",
    command_setup_open="MNMEN ON",
    command_setup_close="MNMEN OFF",
    command_setup_query="MNMEN?",
    command_channel_level_adjust="MNCHL",
    command_dimmer_toggle="DIM SEL",
    command_dimmer_set="DIM {mode}",
    command_channel_volume="CV{channel} {value}",
    command_channel_volumes_reset="CVZRL",
    command_subwoofer_on_off="PSSWR {mode}",
    command_subwoofer_level="PSSWL{number} {mode}",
    command_lfe="PSLFE {mode}",
    command_tactile_transducer="SSTTR {mode}",
    command_tactile_transducer_level="SSTTRLEV {mode}",
    command_tactile_transducer_lpf="SSTTRLPF {frequency}",
    command_delay_up="PSDELAY UP",
    command_delay_down="PSDELAY DOWN",
    command_auromatic_3d_preset="PSAUROPR {preset}",
    command_auromatic_3d_strength="PSAUROST {value}",
    command_auro_3d_mode="PSAUROMODE {mode}",
    command_dirac_filter="PSDIRAC {filter}",
    command_eco_mode="ECO{mode}",
    command_lfc="PSLFC {mode}",
    command_containment_amount="PSCNTAMT {value}",
    command_loudness_management="PSLOM {mode}",
    command_bass_sync="PSBSC {mode}",
    command_dialog_enhancer="PSDEH {level}",
    command_hdmi_output="VSMONI{output}",
    command_hdmi_audio_decode="VSAUDIO {mode}",
    command_quick_select_mode="MSQUICK{number}",
    command_quick_select_memory="MSQUICK{number} MEMORY",
    command_smart_select_mode="MSSMART{number}",  # Marantz
    command_smart_select_memory="MSSMART{number} MEMORY",  # Marantz
    command_auto_standby="STBY{mode}",
    command_sleep="SLP{value}",
    command_center_spread="PSCES {mode}",
    command_video_processing_mode="VSVPM{mode}",
    command_room_size="PSRSZ {size}",
    command_status="RCSHP0230030",
    command_system_reset="SYRST",
    command_network_restart="NSRBT",
    command_trigger="TR{number} {mode}",
    command_speaker_preset="SPPR {number}",
    command_bluetooth_transmitter="BTTX {mode}",
    command_dialog_control="PSDIC {value}",
    command_speaker_virtualizer="PSSPV {mode}",
    command_effect_speaker_selection="PSSP:{mode}",
    command_drc="PSDRC {mode}",
    command_delay_time="PSDEL {value}",
    command_audio_restorer="PSRSTR {mode}",
    command_remote_control_lock="SYREMOTE LOCK {mode}",
    command_panel_lock="SYPANEL LOCK {mode}",
    command_panel_and_volume_lock="SYPANEL+V LOCK ON",
    command_graphic_eq="PSGEQ {mode}",
    command_headphone_eq="PSHEQ {mode}",
)

ZONE2_TELNET_COMMANDS = TelnetCommands(
    command_sel_src="Z2",
    command_fav_src="Z2",
    command_power_on="Z2ON",
    command_power_standby="Z2OFF",
    command_volume_up="Z2UP",
    command_volume_down="Z2DOWN",
    command_set_volume="Z2{volume:02d}",
    command_mute_on="Z2MUON",
    command_mute_off="Z2MUOFF",
    command_sel_sound_mode="MS",
    command_neural_x_on_off="PSNEURAL {mode}",
    command_imax_auto_off="PSIMAX {mode}",
    command_imax_audio_settings="PSIMAXAUD {mode}",
    command_imax_hpf="PSIMAXHPF {frequency}",
    command_imax_lpf="PSIMAXLPF {frequency}",
    command_imax_subwoofer_mode="PSIMAXSWM {mode}",
    command_imax_subwoofer_output="PSIMAXSWO {mode}",
    command_cinema_eq="PSCINEMA EQ.{mode}",
    command_set_all_zone_stereo="MN",
    command_pause="NS9B",
    command_play="NS9A",
    command_multieq="PSMULTEQ:",
    command_dynamiceq="PSDYNEQ ",
    command_reflevoffset="PSREFLEV ",
    command_dynamicvol="PSDYNVOL ",
    command_tonecontrol="PSTONE CTRL ",
    command_bass="PSBAS ",
    command_treble="PSTRE ",
    command_cusor_up="MNCUP",
    command_cusor_down="MNCDN",
    command_cusor_left="MNCLT",
    command_cusor_right="MNCRT",
    command_cusor_enter="MNENT",
    command_back="MNRTN",
    command_info="MNINF",
    command_options="MNOPT",
    command_setup_open="MNMEN ON",
    command_setup_close="MNMEN OFF",
    command_setup_query="MNMEN?",
    command_channel_level_adjust="MNCHL",
    command_dimmer_toggle="DIM SEL",
    command_dimmer_set="DIM {mode}",
    command_channel_volume="CV{channel} {value}",
    command_channel_volumes_reset="CVZRL",
    command_subwoofer_on_off="PSSWR {mode}",
    command_subwoofer_level="PSSWL{number} {mode}",
    command_lfe="PSLFE {mode}",
    command_tactile_transducer="SSTTR {mode}",
    command_tactile_transducer_level="SSTTRLEV {mode}",
    command_tactile_transducer_lpf="SSTTRLPF {frequency}",
    command_delay_up="PSDELAY UP",
    command_delay_down="PSDELAY DOWN",
    command_auromatic_3d_preset="PSAUROPR {preset}",
    command_auromatic_3d_strength="PSAUROST {value}",
    command_auro_3d_mode="PSAUROMODE {mode}",
    command_dirac_filter="PSDIRAC {filter}",
    command_eco_mode="ECO{mode}",
    command_lfc="PSLFC {mode}",
    command_containment_amount="PSCNTAMT {value}",
    command_loudness_management="PSLOM {mode}",
    command_bass_sync="PSBSC {mode}",
    command_dialog_enhancer="PSDEH {level}",
    command_hdmi_output="VSMONI{output}",
    command_hdmi_audio_decode="VSAUDIO {mode}",
    command_quick_select_mode="MSQUICK{number}",
    command_quick_select_memory="MSQUICK{number} MEMORY",
    command_smart_select_mode="MSSMART{number}",  # Marantz
    command_smart_select_memory="MSSMART{number} MEMORY",  # Marantz
    command_auto_standby="STBY{mode}",
    command_sleep="SLP{value}",
    command_center_spread="PSCES {mode}",
    command_video_processing_mode="VSVPM{mode}",
    command_room_size="PSRSZ {size}",
    command_status="RCSHP0230030",
    command_system_reset="SYRST",
    command_network_restart="NSRBT",
    command_trigger="TR{number} {mode}",
    command_speaker_preset="SPPR {number}",
    command_bluetooth_transmitter="BTTX {mode}",
    command_dialog_control="PSDIC {value}",
    command_speaker_virtualizer="PSSPV {mode}",
    command_effect_speaker_selection="PSSP:{mode}",
    command_drc="PSDRC {mode}",
    command_delay_time="PSDEL {value}",
    command_audio_restorer="PSRSTR {mode}",
    command_remote_control_lock="SYREMOTE LOCK {mode}",
    command_panel_lock="SYPANEL LOCK {mode}",
    command_panel_and_volume_lock="SYPANEL+V LOCK ON",
    command_graphic_eq="PSGEQ {mode}",
    command_headphone_eq="PSHEQ {mode}",
)

ZONE3_TELNET_COMMANDS = TelnetCommands(
    command_sel_src="Z3",
    command_fav_src="Z3",
    command_power_on="Z3ON",
    command_power_standby="Z3OFF",
    command_volume_up="Z3UP",
    command_volume_down="Z3DOWN",
    command_set_volume="Z3{volume:02d}",
    command_mute_on="Z3MUON",
    command_mute_off="Z3MUOFF",
    command_sel_sound_mode="MS",
    command_neural_x_on_off="PSNEURAL {mode}",
    command_imax_auto_off="PSIMAX {mode}",
    command_imax_audio_settings="PSIMAXAUD {mode}",
    command_imax_hpf="PSIMAXHPF {frequency}",
    command_imax_lpf="PSIMAXLPF {frequency}",
    command_imax_subwoofer_mode="PSIMAXSWM {mode}",
    command_imax_subwoofer_output="PSIMAXSWO {mode}",
    command_cinema_eq="PSCINEMA EQ.{mode}",
    command_set_all_zone_stereo="MN",
    command_pause="NS9B",
    command_play="NS9A",
    command_multieq="PSMULTEQ:",
    command_dynamiceq="PSDYNEQ ",
    command_reflevoffset="PSREFLEV ",
    command_dynamicvol="PSDYNVOL ",
    command_tonecontrol="PSTONE CTRL ",
    command_bass="PSBAS ",
    command_treble="PSTRE ",
    command_cusor_up="MNCUP",
    command_cusor_down="MNCDN",
    command_cusor_left="MNCLT",
    command_cusor_right="MNCRT",
    command_cusor_enter="MNENT",
    command_back="MNRTN",
    command_info="MNINF",
    command_options="MNOPT",
    command_setup_open="MNMEN ON",
    command_setup_close="MNMEN OFF",
    command_setup_query="MNMEN?",
    command_channel_level_adjust="MNCHL",
    command_dimmer_toggle="DIM SEL",
    command_dimmer_set="DIM {mode}",
    command_channel_volume="CV{channel} {value}",
    command_channel_volumes_reset="CVZRL",
    command_subwoofer_on_off="PSSWR {mode}",
    command_subwoofer_level="PSSWL{number} {mode}",
    command_lfe="PSLFE {mode}",
    command_tactile_transducer="SSTTR {mode}",
    command_tactile_transducer_level="SSTTRLEV {mode}",
    command_tactile_transducer_lpf="SSTTRLPF {frequency}",
    command_delay_up="PSDELAY UP",
    command_delay_down="PSDELAY DOWN",
    command_auromatic_3d_preset="PSAUROPR {preset}",
    command_auromatic_3d_strength="PSAUROST {value}",
    command_auro_3d_mode="PSAUROMODE {mode}",
    command_dirac_filter="PSDIRAC {filter}",
    command_eco_mode="ECO{mode}",
    command_lfc="PSLFC {mode}",
    command_containment_amount="PSCNTAMT {value}",
    command_loudness_management="PSLOM {mode}",
    command_bass_sync="PSBSC {mode}",
    command_dialog_enhancer="PSDEH {level}",
    command_hdmi_output="VSMONI{output}",
    command_hdmi_audio_decode="VSAUDIO {mode}",
    command_quick_select_mode="MSQUICK{number}",
    command_quick_select_memory="MSQUICK{number} MEMORY",
    command_smart_select_mode="MSSMART{number}",  # Marantz
    command_smart_select_memory="MSSMART{number} MEMORY",  # Marantz
    command_auto_standby="STBY{mode}",
    command_sleep="SLP{value}",
    command_center_spread="PSCES {mode}",
    command_video_processing_mode="VSVPM{mode}",
    command_room_size="PSRSZ {size}",
    command_status="RCSHP0230030",
    command_system_reset="SYRST",
    command_network_restart="NSRBT",
    command_trigger="TR{number} {mode}",
    command_speaker_preset="SPPR {number}",
    command_bluetooth_transmitter="BTTX {mode}",
    command_dialog_control="PSDIC {value}",
    command_speaker_virtualizer="PSSPV {mode}",
    command_effect_speaker_selection="PSSP:{mode}",
    command_drc="PSDRC {mode}",
    command_delay_time="PSDEL {value}",
    command_audio_restorer="PSRSTR {mode}",
    command_remote_control_lock="SYREMOTE LOCK {mode}",
    command_panel_lock="SYPANEL LOCK {mode}",
    command_panel_and_volume_lock="SYPANEL+V LOCK ON",
    command_graphic_eq="PSGEQ {mode}",
    command_headphone_eq="PSHEQ {mode}",
)

# States
POWER_ON = "ON"
POWER_OFF = "OFF"
POWER_STANDBY = "STANDBY"
POWER_STATES = {POWER_ON, POWER_OFF, POWER_STANDBY}
STATE_ON = "on"
STATE_OFF = "off"
STATE_PLAYING = "playing"
STATE_PAUSED = "paused"
SETTINGS_MENU_ON = "ON"
SETTINGS_MENU_OFF = "OFF"
SETTINGS_MENU_STATES = {SETTINGS_MENU_ON, SETTINGS_MENU_OFF}
DIMER_BRIGHT = "BRI"
DIMER_DIM = "DIM"
DIMER_DARK = "DAR"
DIMER_OFF = "OFF"
DIMMER_STATES = {DIMER_BRIGHT, DIMER_DIM, DIMER_DARK, DIMER_OFF}

# Zones
ALL_ZONES = "All"
MAIN_ZONE = "Main"
ZONE2 = "Zone2"
ZONE3 = "Zone3"
VALID_ZONES = {MAIN_ZONE, ZONE2, ZONE3}

# Setup additional zones
NO_ZONES = None
ZONE2_NAME = {ZONE2: None}
ZONE3_NAME = {ZONE3: None}
ZONE2_ZONE3_NAME = {ZONE2: None, ZONE3: None}

# AppCommand related
APPCOMMAND_CMD_TEXT = "cmd_text"
APPCOMMAND_NAME = "name"

# Audyssey parameter
MULTI_EQ_MAP_APPCOMMAND = {"0": "Off", "1": "Flat", "2": "L/R Bypass", "3": "Reference"}
MULTI_EQ_MAP_TELNET = {
    "OFF": "Off",
    "FLAT": "Flat",
    "BYP.LR": "L/R Bypass",
    "AUDYSSEY": "Reference",
    "MANUAL": "Manual",
}
MULTI_EQ_MAP = {**MULTI_EQ_MAP_APPCOMMAND, **MULTI_EQ_MAP_TELNET}
MULTI_EQ_MAP_LABELS_APPCOMMAND = {
    value: key for key, value in MULTI_EQ_MAP_APPCOMMAND.items()
}
MULTI_EQ_MAP_LABELS_TELNET = {value: key for key, value in MULTI_EQ_MAP_TELNET.items()}

REF_LVL_OFFSET_MAP_APPCOMMAND = {"0": "0dB", "1": "+5dB", "2": "+10dB", "3": "+15dB"}
REF_LVL_OFFSET_MAP_TELNET = {"0": "0dB", "5": "+5dB", "10": "+10dB", "15": "+15dB"}
REF_LVL_OFFSET_MAP = {**REF_LVL_OFFSET_MAP_APPCOMMAND, **REF_LVL_OFFSET_MAP_TELNET}
REF_LVL_OFFSET_MAP_LABELS_APPCOMMAND = {
    value: key for key, value in REF_LVL_OFFSET_MAP_APPCOMMAND.items()
}
REF_LVL_OFFSET_MAP_LABELS_TELNET = {
    value: key for key, value in REF_LVL_OFFSET_MAP_TELNET.items()
}

DYNAMIC_VOLUME_MAP_APPCOMMAND = {"0": "Off", "1": "Light", "2": "Medium", "3": "Heavy"}
DYNAMIC_VOLUME_MAP_TELNET = {
    "OFF": "Off",
    "LIT": "Light",
    "MED": "Medium",
    "HEV": "Heavy",
}
DYNAMIC_VOLUME_MAP = {**DYNAMIC_VOLUME_MAP_APPCOMMAND, **DYNAMIC_VOLUME_MAP_TELNET}
DYNAMIC_VOLUME_MAP_LABELS_APPCOMMAND = {
    value: key for key, value in DYNAMIC_VOLUME_MAP_APPCOMMAND.items()
}
DYNAMIC_VOLUME_MAP_LABELS_TELNET = {
    value: key for key, value in DYNAMIC_VOLUME_MAP_TELNET.items()
}

AudioRestorers = Literal["Off", "Low", "Medium", "High"]
"""Audio Restorer settings."""

AUDIO_RESTORER_MAP = {
    "Off": "OFF",
    "Low": "LOW",
    "Medium": "MED",
    "High": "HI",
}
AUDIO_RESTORER_MAP_LABELS = {value: key for key, value in AUDIO_RESTORER_MAP.items()}

AuroMatic3DPresets = Literal[
    "Small",
    "Medium",
    "Large",
    "Speech",
    "Movie",
]
"""Auro-Matic 3D Presets."""

AURO_MATIC_3D_PRESET_MAP = {
    "Small": "SMA",
    "Medium": "MED",
    "Large": "LAR",
    "Speech": "SPE",
    "Movie": "MOV",
}
AURO_MATIC_3D_PRESET_MAP_LABELS = {
    value: key for key, value in AURO_MATIC_3D_PRESET_MAP.items()
}

Auro3DModes = Literal["Direct", "Channel Expansion"]
"""Auro 3D Modes."""

AURO_3D_MODE_MAP = {"Direct": "DRCT", "Channel Expansion": "EXP"}
AURO_3D_MODE_MAP_MAP_LABELS = {value: key for key, value in AURO_3D_MODE_MAP.items()}

AutoStandbys = Literal["OFF", "15M", "30M", "60M"]

BluetoothOutputModes = Literal["Bluetooth + Speakers", "Bluetooth Only"]
"""Bluetooth Output Modes."""

BLUETOOTH_OUTPUT_MODES_MAP = {
    "Bluetooth + Speakers": "SP",
    "Bluetooth Only": "BT",
}
BLUETOOTH_OUTPUT_MAP_LABELS = {
    value: key for key, value in BLUETOOTH_OUTPUT_MODES_MAP.items()
}

DIMMER_MODE_MAP = {
    "Off": DIMER_OFF,
    "Dark": DIMER_DARK,
    "Dim": DIMER_DIM,
    "Bright": DIMER_BRIGHT,
}
DIMMER_MODE_MAP_LABELS = {value: key for key, value in DIMMER_MODE_MAP.items()}

DimmerModes = Literal["Off", "Dark", "Dim", "Bright"]
"""Dimmer Modes."""

DIRAC_FILTER_MAP = {"Off": "OFF", "Slot 1": "1", "Slot 2": "2", "Slot 3": "3"}
DIRAC_FILTER_MAP_LABELS = {value: key for key, value in DIRAC_FILTER_MAP.items()}

DiracFilters = Literal["Slot 1", "Slot 2", "Slot 3", "Off"]
"""Dirac Filters."""

ECO_MODE_MAP = {
    "On": "ON",
    "Auto": "AUTO",
    "Off": "OFF",
}
ECO_MODE_MAP_LABELS = {value: key for key, value in ECO_MODE_MAP.items()}

EcoModes = Literal["On", "Auto", "Off"]
"""Eco Modes."""

EffectSpeakers = Literal[
    "Floor",
    "Front",
    "Front Height",
    "Front Wide",
    "Front Height + Front Wide",
    "Height + Floor",
    "Surround Back",
    "Surround Back + Front Height",
    "Surround Back + Front Wide",
]
"""Effect Speakers."""

EFFECT_SPEAKER_SELECTION_MAP = {
    "Floor": "FL",
    "Front": "FR",
    "Front Height": "FH",
    "Front Wide": "FW",
    "Front Height + Front Wide": "HW",
    "Height + Floor": "HF",
    "Surround Back": "SB",
    "Surround Back + Front Height": "BH",
    "Surround Back + Front Wide": "BW",
}
EFFECT_SPEAKER_SELECTION_MAP_LABELS = {
    value: key for key, value in EFFECT_SPEAKER_SELECTION_MAP.items()
}

DRCs = Literal["AUTO", "LOW", "MID", "HI", "OFF"]
"""Dynamic Range Control (DRC) Settings."""

HDMI_OUTPUT_MAP = {
    "Auto": "AUTO",
    "HDMI1": "1",
    "HDMI2": "2",
}
HDMI_OUTPUT_MAP_LABELS = {
    "MONIAUTO": "Auto",
    "MONI1": "HDMI1",
    "MONI2": "HDMI2",
}

HDMIOutputs = Literal["Auto", "HDMI1", "HDMI2"]
"""HDMI Output Modes."""

HDMIAudioDecodes = Literal["AMP", "TV"]

Subwoofers = Literal["Subwoofer", "Subwoofer 2", "Subwoofer 3", "Subwoofer 4"]
"""Subwoofers."""

SUBWOOFERS_MAP = {
    "Subwoofer": "",
    "Subwoofer 2": "2",
    "Subwoofer 3": "3",
    "Subwoofer 4": "4",
}
"""Subwoofers."""

SUBWOOFERS_MAP_LABELS = {
    "SWL": "Subwoofer",
    "SWL2": "Subwoofer 2",
    "SWL3": "Subwoofer 3",
    "SWL4": "Subwoofer 4",
}

CHANNEL_MAP = {
    "Front Left": "FL",
    "Front Right": "FR",
    "Center": "C",
    "Subwoofer": "SW",
    "Subwoofer 2": "SW2",
    "Subwoofer 3": "SW3",
    "Subwoofer 4": "SW4",
    "Surround Left": "SL",
    "Surround Right": "SR",
    "Surround Back Left": "SBL",
    "Surround Back Right": "SBR",
    "Front Height Left": "FHL",
    "Front Height Right": "FHR",
    "Front Wide Left": "FWL",
    "Front Wide Right": "FWR",
    "Top Front Left": "TFL",
    "Top Front Right": "TFR",
    "Top Middle Left": "TML",
    "Top Middle Right": "TMR",
    "Top Rear Left": "TRL",
    "Top Rear Right": "TRR",
    "Rear Height Left": "RHL",
    "Rear Height Right": "RHR",
    "Front Dolby Left": "FDL",
    "Front Dolby Right": "FDR",
    "Surround Dolby Left": "SDL",
    "Surround Dolby Right": "SDR",
    "Back Dolby Left": "BDL",
    "Back Dolby Right": "BDR",
    "Surround Height Left": "SHL",
    "Surround Height Right": "SHR",
    "Top Surround": "TS",
    "Center Height": "CH",
}
CHANNEL_MAP_LABELS = {value: key for key, value in CHANNEL_MAP.items()}

Channels = Literal[
    "Front Left",
    "Front Right",
    "Center",
    "Subwoofer",
    "Subwoofer 2",
    "Subwoofer 3",
    "Subwoofer 4",
    "Surround Left",
    "Surround Right",
    "Surround Back Left",
    "Surround Back Right",
    "Front Height Left",
    "Front Height Right",
    "Front Wide Left",
    "Front Wide Right",
    "Top Front Left",
    "Top Front Right",
    "Top Middle Left",
    "Top Middle Right",
    "Top Rear Left",
    "Top Rear Right",
    "Rear Height Left",
    "Rear Height Right",
    "Front Dolby Left",
    "Front Dolby Right",
    "Surround Dolby Left",
    "Surround Dolby Right",
    "Back Dolby Left",
    "Back Dolby Right",
    "Surround Height Left",
    "Surround Height Right",
    "Top Surround",
    "Center Height",
]
"""Receiver Channels."""

CHANNEL_VOLUME_MAP = {
    "38": -12.0,
    "385": -11.5,
    "39": -11.0,
    "395": -10.5,
    "40": -10.0,
    "405": -9.5,
    "41": -9.0,
    "415": -8.5,
    "42": -8.0,
    "425": -7.5,
    "43": -7.0,
    "435": -6.5,
    "44": -6.0,
    "445": -5.5,
    "45": -5.0,
    "455": -4.5,
    "46": -4.0,
    "465": -3.5,
    "47": -3.0,
    "475": -2.5,
    "48": -2.0,
    "485": -1.5,
    "49": -1.0,
    "495": -0.5,
    "50": 0.0,
    "505": 0.5,
    "51": 1.0,
    "515": 1.5,
    "52": 2.0,
    "525": 2.5,
    "53": 3.0,
    "535": 3.5,
    "54": 4.0,
    "545": 4.5,
    "55": 5.0,
    "555": 5.5,
    "56": 6.0,
    "565": 6.5,
    "57": 7.0,
    "575": 7.5,
    "58": 8.0,
    "585": 8.5,
    "59": 9.0,
    "595": 9.5,
    "60": 10.0,
    "605": 10.5,
    "61": 11.0,
    "615": 11.5,
    "62": 12.0,
}
CHANNEL_VOLUME_MAP_LABELS = {value: key for key, value in CHANNEL_VOLUME_MAP.items()}

DialogEnhancerLevels = Literal["Off", "Low", "Medium", "High"]
"""Dialog Enhancer Levels."""

DIALOG_ENHANCER_LEVEL_MAP = {
    "Off": "OFF",
    "Low": "LOW",
    "Medium": "MED",
    "High": "HIGH",
}

DIALOG_ENHANCER_LEVEL_MAP_LABELS = {
    value: key for key, value in DIALOG_ENHANCER_LEVEL_MAP.items()
}

TransducerLPFs = Literal[
    "40 Hz",
    "60 Hz",
    "80 Hz",
    "90 Hz",
    "100 Hz",
    "110 Hz",
    "120 Hz",
    "150 Hz",
    "180 Hz",
    "200 Hz",
    "250 Hz",
]
"""Tactile Transducer Low Pass Frequencies."""

IMAXHPFs = Literal[
    "40", "60", "80", "90", "100", "110", "120", "150", "180", "200", "250"
]
"""IMAX High Pass Frequencies."""

IMAXLPFs = Literal["80", "90", "100", "110", "120", "150", "180", "200", "250"]
"""IMAX Low Pass Frequencies."""

PanelLocks = Literal["Panel", "Panel + Master Volume"]
"""Panel Lock Modes."""

RoomSizes = Literal[
    "S",
    "MS",
    "M",
    "ML",
    "L",
]
"""Room Sizes."""

VideoProcessingModes = Literal["Auto", "Game", "Movie", "Bypass"]
"""Video Processing Modes."""

VIDEO_PROCESSING_MODES_MAP = {
    "Auto": "AUTO",
    "Game": "GAME",
    "Movie": "MOVI",
    "Bypass": "BYP",
}

VIDEO_PROCESSING_MODES_MAP_LABELS = {
    value: key for key, value in VIDEO_PROCESSING_MODES_MAP.items()
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the REST and Telnet APIs to Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import asyncio
import contextlib
import logging
import sys
import time
import xml.etree.ElementTree as ET
from collections import defaultdict
from collections.abc import Hashable
from io import BytesIO
from typing import (
    Awaitable,
    Callable,
    Coroutine,
    DefaultDict,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    cast,
)

import attr
import httpx
from defusedxml import DefusedXmlException
from defusedxml.ElementTree import ParseError, fromstring

from .appcommand import AppCommandCmd
from .const import (
    ALL_TELNET_EVENTS,
    ALL_ZONE_TELNET_EVENTS,
    ALL_ZONES,
    APPCOMMAND0300_URL,
    APPCOMMAND_CMD_TEXT,
    APPCOMMAND_NAME,
    APPCOMMAND_URL,
    DENON_ATTR_SETATTR,
    MAIN_ZONE,
    TELNET_EVENTS,
    TELNET_SOURCES,
    ZONE2,
    ZONE3,
)
from .decorators import async_handle_receiver_exceptions, cache_result
from .exceptions import (
    AvrIncompleteResponseError,
    AvrInvalidResponseError,
    AvrNetworkError,
    AvrProcessingError,
    AvrTimoutError,
)

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout
else:
    from asyncio import timeout as asyncio_timeout

_LOGGER = logging.getLogger(__name__)

_MONITOR_INTERVAL = 30


def get_default_async_client() -> httpx.AsyncClient:
    """Get the default httpx.AsyncClient."""
    return httpx.AsyncClient()


def telnet_event_map_factory() -> Dict[str, List]:
    """Create telnet event map."""
    event_map: DefaultDict[str, List] = defaultdict(list)
    for event in TELNET_EVENTS:
        event_map[event[0:2]].append(event)
    for value in event_map.values():
        value.sort(key=len, reverse=True)
    return dict(event_map)


@attr.s(auto_attribs=True, hash=False)
class HTTPXAsyncClient:
    """Perform cached HTTP calls with httpx.AsyncClient."""

    client_getter: Callable[[], httpx.AsyncClient] = attr.ib(
        validator=attr.validators.is_callable(),
        default=get_default_async_client,
        init=False,
    )

    def __hash__(self) -> int:
        """Hash the class using its ID that caching works."""
        return id(self)

    @cache_result
    @async_handle_receiver_exceptions
    async def async_get(
        self,
        url: str,
        timeout: float,
        read_timeout: float,
        *,
        cache_id: Hashable = None,
    ) -> httpx.Response:
        """Call GET endpoint of Denon AVR receiver asynchronously."""
        client = self.client_getter()
        try:
            res = await client.get(
                url, timeout=httpx.Timeout(timeout, read=read_timeout)
            )
            res.raise_for_status()
        finally:
            # Close the default AsyncClient but keep custom clients open
            if self.is_default_async_client():
                await client.aclose()

        return res

    @cache_result
    @async_handle_receiver_exceptions
    async def async_post(
        self,
        url: str,
        timeout: float,
        read_timeout: float,
        *,
        content: Optional[bytes] = None,
        data: Optional[Dict] = None,
        cache_id: Hashable = None,
    ) -> httpx.Response:
        """Call GET endpoint of Denon AVR receiver asynchronously."""
        client = self.client_getter()
        try:
            res = await client.post(
                url,
                content=content,
                data=data,
                timeout=httpx.Timeout(timeout, read=read_timeout),
            )
            res.raise_for_status()
        finally:
            # Close the default AsyncClient but keep custom clients open
            if self.is_default_async_client():
                await client.aclose()

        return res

    def is_default_async_client(self) -> bool:
        """Check if default httpx.AsyncClient getter is used."""
        return self.client_getter is get_default_async_client


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRApi:
    """Perform API calls to Denon AVR REST interface."""

    host: str = attr.ib(converter=str, default="localhost")
    port: int = attr.ib(converter=int, default=80)
    timeout: float = attr.ib(converter=float, default=2.0)
    read_timeout: float = attr.ib(converter=float, default=15.0)
    _appcommand_update_tags: Tuple[AppCommandCmd] = attr.ib(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(AppCommandCmd),
            attr.validators.instance_of(tuple),
        ),
        default=attr.Factory(tuple),
    )
    _appcommand0300_update_tags: Tuple[AppCommandCmd] = attr.ib(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(AppCommandCmd),
            attr.validators.instance_of(tuple),
        ),
        default=attr.Factory(tuple),
    )
    httpx_async_client: HTTPXAsyncClient = attr.ib(
        validator=attr.validators.instance_of(HTTPXAsyncClient),
        default=attr.Factory(HTTPXAsyncClient),
        init=False,
    )

    async def async_get(
        self,
        request: str,
        *,
        port: Optional[int] = None,
        cache_id: Hashable = None,
    ) -> httpx.Response:
        """Call GET endpoint of Denon AVR receiver asynchronously."""
        # Use default port of the receiver if no different port is specified
        port = port if port is not None else self.port

        endpoint = f"http://{self.host}:{port}{request}"

        return await self.httpx_async_client.async_get(
            endpoint, self.timeout, self.read_timeout, cache_id=cache_id
        )

    async def async_post(
        self,
        request: str,
        *,
        content: Optional[bytes] = None,
        data: Optional[Dict] = None,
        port: Optional[int] = None,
        cache_id: Hashable = None,
    ) -> httpx.Response:
        """Call POST endpoint of Denon AVR receiver asynchronously."""
        # Use default port of the receiver if no different port is specified
        port = port if port is not None else self.port

        endpoint = f"http://{self.host}:{port}{request}"

        return await self.httpx_async_client.async_post(
            endpoint,
            self.timeout,
            self.read_timeout,
            content=content,
            data=data,
            cache_id=cache_id,
        )

    async def async_get_command(self, request: str) -> str:
        """Send HTTP GET command to Denon AVR receiver asynchronously."""
        # HTTP GET to endpoint
        res = await self.async_get(request)
        # Return text
        return res.text

    async def async_get_xml(
        self, request: str, *, cache_id: Hashable = None
    ) -> ET.Element:
        """Return XML data from HTTP GET endpoint asynchronously."""
        # HTTP GET to endpoint
        res = await self.async_get(request, cache_id=cache_id)
        # create ElementTree
        try:
            xml_root = fromstring(res.text)
        except (
            ET.ParseError,
            DefusedXmlException,
            ParseError,
            UnicodeDecodeError,
        ) as err:
            raise AvrInvalidResponseError(f"XMLParseError: {err}", request) from err
        # Check validity of XML
        self.check_xml_validity(request, xml_root)
        # Return ElementTree element
        return xml_root

    async def async_post_appcommand(
        self, request: str, cmds: Tuple[AppCommandCmd], *, cache_id: Hashable = None
    ) -> ET.Element:
        """Return XML from Appcommand(0300) endpoint asynchronously."""
        # Prepare XML body for POST call
        content = self.prepare_appcommand_body(cmds)
        _LOGGER.debug("Content for %s endpoint: %s", request, content)
        # HTTP POST to endpoint
        res = await self.async_post(request, content=content, cache_id=cache_id)
        # create ElementTree
        try:
            xml_root = fromstring(res.text)
        except (
            ET.ParseError,
            DefusedXmlException,
            ParseError,
            UnicodeDecodeError,
        ) as err:
            raise AvrInvalidResponseError(f"XMLParseError: {err}", request) from err
        # Check validity of XML
        self.check_xml_validity(request, xml_root)
        # Add query tags to result
        xml_root = self.add_query_tags_to_result(request, cmds, xml_root)
        # Return ElementTree element
        return xml_root

    def add_appcommand_update_tag(self, tag: AppCommandCmd) -> None:
        """Add appcommand tag for full update."""
        if tag.cmd_id != "1":
            raise ValueError(f"cmd_id is {tag.cmd_id} but must be 1")

        # Remove response pattern from tag because it is not relevant for query
        tag = attr.evolve(tag, response_pattern=tuple())

        if tag not in self._appcommand_update_tags:
            _LOGGER.debug("Add tag %s to AppCommand update tuple", tag)
            self._appcommand_update_tags = (*self._appcommand_update_tags, tag)

    def add_appcommand0300_update_tag(self, tag: AppCommandCmd) -> None:
        """Add appcommand0300 tag for full update."""
        if tag.cmd_id != "3":
            raise ValueError(f"cmd_id is {tag.cmd_id} but must be 3")

        # Remove response pattern from tag because it is not relevant for query
        tag = attr.evolve(tag, response_pattern=tuple())

        if tag not in self._appcommand0300_update_tags:
            _LOGGER.debug("Add tag %s to AppCommand0300 update tuple", tag)
            self._appcommand0300_update_tags = (*self._appcommand0300_update_tags, tag)

    async def async_get_global_appcommand(
        self, appcommand0300: bool = False, cache_id: Optional[Hashable] = None
    ) -> ET.Element:
        """Get XML structure for full AppCommand update."""
        if appcommand0300:
            tags = self._appcommand0300_update_tags
            url = APPCOMMAND0300_URL
        else:
            tags = self._appcommand_update_tags
            url = APPCOMMAND_URL

        res = await self.async_post_appcommand(url, tags, cache_id=cache_id)
        return res

    @staticmethod
    def add_query_tags_to_result(
        request: str, cmd_list: Tuple[AppCommandCmd], xml_root: ET.Element
    ) -> ET.Element:
        """
        Add query tags to corresponding elements of result XML.

        This is used to identitfy the result tags.
        """
        if len(cmd_list) != len(xml_root):
            raise AvrIncompleteResponseError(
                (
                    "Invalid length of response XML. Query has"
                    f" {len(cmd_list)} elements, response {len(xml_root)}"
                ),
                request,
            )

        for i, child in enumerate(xml_root):
            if child.tag not in ["cmd", "error"]:
                raise AvrInvalidResponseError(
                    (
                        'Returned document contains a tag other than "cmd" and'
                        f' "error": {child.tag}'
                    ),
                    request,
                )
            # Find corresponding attributes from request XML if set and add
            # tag to cmd element
            if cmd_list[i].cmd_text is not None:
                child.set(APPCOMMAND_CMD_TEXT, cmd_list[i].cmd_text)
            if cmd_list[i].name is not None:
                child.set(APPCOMMAND_NAME, cmd_list[i].name)

        return xml_root

    @staticmethod
    def check_xml_validity(request: str, xml_root: ET.Element) -> None:
        """Check if there is a valid Denon XML and not a HTML page."""
        if xml_root.tag == "html":
            raise AvrInvalidResponseError("Returned document contains HTML", request)

    @staticmethod
    def prepare_appcommand_body(cmd_list: Tuple[AppCommandCmd]) -> bytes:
        """Prepare HTTP POST body to AppCommand(0300).xml end point."""
        # Buffer XML body as binary IO
        body = BytesIO()

        # Denon AppCommand.xml acts weird. It returns an error when the tx
        # element consists of more than 5 cmd elements, but it accepts
        # multiple XML root elements
        chunks = [cmd_list[i : i + 5] for i in range(0, len(cmd_list), 5)]

        for i, chunk in enumerate(chunks):
            # Prepare POST XML body for AppCommand.xml
            xml_root = ET.Element("tx")

            for cmd in chunk:
                # Append tags for each element
                cmd_element = ET.Element("cmd")
                cmd_element.set("id", cmd.cmd_id)

                # Optional cmd text
                cmd_element.text = cmd.cmd_text

                # Optional name tag
                if cmd.name is not None:
                    name_element = ET.Element("name")
                    name_element.text = cmd.name
                    cmd_element.append(name_element)

                # Optional list tag
                if cmd.param_list is not None:
                    param_list_element = ET.Element("list")
                    for param in cmd.param_list:
                        param_element = ET.Element("param")
                        param_element.set("name", param.name)
                        param_element.text = param.text
                        param_list_element.append(param_element)
                    cmd_element.append(param_list_element)

                xml_root.append(cmd_element)

                # Optional command parameter
                if cmd.set_command is not None:
                    command_element = ET.Element(cmd.set_command.name)
                    command_element.text = cmd.set_command.text
                    xml_root.append(command_element)

            xml_tree = ET.ElementTree(xml_root)
            # XML declaration only for the first chunk
            xml_tree.write(body, encoding="utf-8", xml_declaration=bool(i == 0))

        body_bytes = body.getvalue()

        # Buffered XML not needed anymore: close
        body.close()

        return body_bytes


class DenonAVRTelnetProtocol(asyncio.Protocol):
    """Protocol for the Denon AVR Telnet interface."""

    def __init__(
        self, on_message: Callable[[str], None], on_connection_lost: Callable[[], None]
    ) -> None:
        """Initialize the protocol."""
        self._buffer = b""
        self.transport: Optional[asyncio.Transport] = None
        self._on_message = on_message
        self._on_connection_lost = on_connection_lost

    @property
    def connected(self) -> bool:
        """Return True if transport is connected."""
        if self.transport is None:
            return False
        return not self.transport.is_closing()

    def write(self, data: str) -> None:
        """Write data to the transport."""
        if self.transport is None:
            return
        if self.transport.is_closing():
            return
        self.transport.write(data.encode("utf-8"))

    def close(self) -> None:
        """Close the connection."""
        if self.transport is not None:
            self.transport.close()

    def data_received(self, data: bytes) -> None:
        """Handle data received."""
        self._buffer += data
        while b"\r" in self._buffer:
            line, _, self._buffer = self._buffer.partition(b"\r")
            with contextlib.suppress(UnicodeDecodeError):
                self._on_message(line.decode("utf-8"))

    def connection_made(self, transport: asyncio.Transport) -> None:
        """Handle connection made."""
        self.transport = transport

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Handle connection lost."""
        self._on_connection_lost()
        super().connection_lost(exc)
        self.transport = None


@attr.s(auto_attribs=True, hash=False, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRTelnetApi:
    """Handle Telnet responses from the Denon AVR Telnet interface."""

    host: str = attr.ib(converter=str, default="localhost")
    timeout: float = attr.ib(converter=float, default=2.0)
    _connection_enabled: bool = attr.ib(default=False)
    _last_message_time: float = attr.ib(default=-1.0)
    _connect_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
    _reconnect_task: asyncio.Task = attr.ib(default=None)
    _monitor_handle: asyncio.TimerHandle = attr.ib(default=None)
    _protocol: DenonAVRTelnetProtocol = attr.ib(default=None)
    _telnet_event_map: Dict[str, List] = attr.ib(
        default=attr.Factory(telnet_event_map_factory)
    )
    _callback_tasks: Set[asyncio.Task] = attr.ib(attr.Factory(set))
    _send_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
    _send_confirmation_timeout: float = attr.ib(converter=float, default=2.0)
    _send_confirmation_event: asyncio.Event = attr.ib(
        default=attr.Factory(asyncio.Event)
    )
    _send_confirmation_command: str = attr.ib(converter=str, default="")
    _send_tasks: Set[asyncio.Task] = attr.ib(attr.Factory(set))
    _callbacks: Dict[str, List[Coroutine]] = attr.ib(
        validator=attr.validators.instance_of(dict),
        default=attr.Factory(dict),
        init=False,
    )
    _raw_callbacks: List[Coroutine] = attr.ib(
        validator=attr.validators.instance_of(list),
        default=attr.Factory(list),
        init=False,
    )

    def __attrs_post_init__(self) -> None:
        """Initialize special attributes."""
        self._register_raw_callback(self._async_send_confirmation_callback)

    async def async_connect(self) -> None:
        """Connect to the receiver asynchronously."""
        _LOGGER.debug("%s: telnet connecting", self.host)
        async with self._connect_lock:
            if self.connected:
                return
            await self._async_establish_connection()

    async def _async_establish_connection(self) -> None:
        """Establish a connection to the receiver."""
        loop = asyncio.get_event_loop()
        _LOGGER.debug("%s: establishing telnet connection", self.host)
        try:
            async with asyncio_timeout(self.timeout):
                transport_protocol = await loop.create_connection(
                    lambda: DenonAVRTelnetProtocol(
                        on_connection_lost=self._handle_disconnected,
                        on_message=self._process_event,
                    ),
                    self.host,
                    23,
                )
        except asyncio.TimeoutError as err:
            _LOGGER.debug("%s: Timeout exception on telnet connect", self.host)
            raise AvrTimoutError(f"TimeoutException: {err}", "telnet connect") from err
        except ConnectionRefusedError as err:
            _LOGGER.debug(
                "%s: Connection refused on telnet connect: %s", self.host, err
            )
            raise AvrNetworkError(
                f"ConnectionRefusedError: {err}", "telnet connect"
            ) from err
        except (OSError, IOError) as err:
            _LOGGER.debug(
                "%s: Connection failed on telnet reconnect: %s", self.host, err
            )
            raise AvrNetworkError(f"OSError: {err}", "telnet connect") from err
        self._protocol = cast(DenonAVRTelnetProtocol, transport_protocol[1])
        _LOGGER.debug("%s: telnet connection established", self.host)
        self._connection_enabled = True
        self._last_message_time = time.monotonic()
        self._schedule_monitor()
        # Trigger update of all attributes
        await self.async_send_commands(
            "ZM?",
            "SI?",
            "MV?",
            "MU?",
            "Z2?",
            "Z2MU?",
            "Z3?",
            "Z3MU?",
            "PSTONE CTRL ?",
            "PSBAS ?",
            "PSTRE ?",
            "PSDYNEQ ?",
            "PSLFC ?",
            "PSCNTAMT ?",
            "PSMULTEQ: ?",
            "PSREFLEV ?",
            "PSDYNVOL ?",
            "MS?",
            "MNMEN?",
            "DIM ?",
            "PSDELAY?",
            "ECO?",
            "VSMONI ?",
            "PSDIRAC ?",
            "CV?",
            "PSNEURAL ?",
            "PSIMAX ?",
            "PSIMAXAUD ?",
            "PSIMAXHPF ?",
            "PSIMAXLPF ?",
            "PSIMAXSWM ?",
            "PSIMAXSWO ?",
            "PSSWR ?",
            "PSSWL ?",
            "SSTTR ?",
            "MSQUICK ?",
            "MSSMART ?",
            "STBY?",
            "SLP?",
            "VSAUDIO ?",
            "PSCES ?",
            "VSVPM ?",
            "PSLFE ?",
            "PSLOM ?",
            "PSBSC ?",
            "PSDEH ?",
            "PSCINEMA EQ. ?",
            "PSAUROPR ?",
            "PSAUROST ?",
            "PSAUROMODE ?",
            "PSRSZ ?",
            "TR?",
            "SPPR ?",
            "BTTX ?",
            "PSDIC ?",
            "PSSPV ?",
            "PSSP: ?",
            "PSDRC ?",
            "PSDEL ?",
            "PSRSTR ?",
            "PSGEQ ?",
            "PSHEQ ?",
            skip_confirmation=True,
        )

    def _schedule_monitor(self) -> None:
        """Start the monitor task."""
        loop = asyncio.get_event_loop()
        self._monitor_handle = loop.call_later(_MONITOR_INTERVAL, self._monitor)

    def _stop_monitor(self) -> None:
        """Stop the monitor task."""
        if self._monitor_handle is not None:
            self._monitor_handle.cancel()
            self._monitor_handle = None

    def _monitor(self) -> None:
        """Monitor the connection."""
        time_since_response = time.monotonic() - self._last_message_time
        if time_since_response > _MONITOR_INTERVAL * 2:
            _LOGGER.info(
                "%s: Keep alive failed, disconnecting and reconnecting", self.host
            )
            self._handle_disconnected()
            return

        if time_since_response > _MONITOR_INTERVAL and self._protocol:
            # Keep the connection alive
            _LOGGER.debug("%s: Sending keep alive", self.host)
            self._protocol.write("PW?\r")
        self._schedule_monitor()

    def _handle_disconnected(self) -> None:
        """Handle disconnected."""
        _LOGGER.debug("%s: handle disconnected", self.host)
        if self._protocol is not None:
            self._protocol.close()
            self._protocol = None
        self._stop_monitor()
        if not self._connection_enabled:
            return
        if self._reconnect_task is None:
            self._reconnect_task = asyncio.create_task(self._async_reconnect())

    async def async_disconnect(self) -> None:
        """Close the connection to the receiver asynchronously."""
        async with self._connect_lock:
            _LOGGER.debug("%s: telnet disconnecting", self.host)
            self._connection_enabled = False
            self._stop_monitor()
            reconnect_task = self._reconnect_task
            if self._reconnect_task is not None:
                self._reconnect_task.cancel()
                self._reconnect_task = None
            if self._protocol is not None:
                self._protocol.close()
                self._protocol = None

            if reconnect_task is not None:
                try:
                    await reconnect_task
                except asyncio.CancelledError:
                    pass
            _LOGGER.debug("%s: telnet disconnected", self.host)

    async def _async_reconnect(self) -> None:
        """Reconnect to the receiver asynchronously."""
        backoff = 0.5

        while self._connection_enabled and not self.healthy:
            async with self._connect_lock:
                _LOGGER.debug("%s: Telnet reconnecting", self.host)
                try:
                    await self._async_establish_connection()
                except AvrTimoutError:
                    _LOGGER.debug(
                        "%s: Timeout exception on telnet reconnect", self.host
                    )
                except AvrNetworkError as err:
                    _LOGGER.debug("%s: %s", self.host, err)
                except AvrProcessingError as err:
                    _LOGGER.debug(
                        "%s: Failed updating state on telnet reconnect: %s",
                        self.host,
                        err,
                    )
                except Exception as err:  # pylint: disable=broad-except
                    _LOGGER.error(
                        "%s: Unexpected exception on telnet reconnect",
                        self.host,
                        exc_info=err,
                    )
                else:
                    _LOGGER.info("%s: Telnet reconnected", self.host)
                    break

            await asyncio.sleep(backoff)
            backoff = min(30.0, backoff * 2)

        self._reconnect_task = None

    def register_callback(
        self, event: str, callback: Callable[[str, str, str], Awaitable[None]]
    ) -> None:
        """Register a callback handler for an event type."""
        # Validate the passed in type
        if event != ALL_TELNET_EVENTS and event not in TELNET_EVENTS:
            raise ValueError(f"{event} is not a valid callback type.")

        if event not in self._callbacks.keys():
            self._callbacks[event] = []
        elif callback in self._callbacks[event]:
            return
        self._callbacks[event].append(callback)

    def unregister_callback(
        self, event: str, callback: Callable[[str, str, str], Awaitable[None]]
    ) -> None:
        """Unregister a callback handler for an event type."""
        if event not in self._callbacks.keys():
            return
        self._callbacks[event].remove(callback)

    def _register_raw_callback(
        self, callback: Callable[[str], Awaitable[None]]
    ) -> None:
        """Register a callback handler for raw telnet messages."""
        if callback in self._raw_callbacks:
            return
        self._raw_callbacks.append(callback)

    def _unregister_raw_callback(
        self, callback: Callable[[str], Awaitable[None]]
    ) -> None:
        """Unregister a callback handler for raw telnet messages."""
        self._raw_callbacks.remove(callback)

    def _process_event(self, message: str) -> None:
        """Process a realtime event."""
        _LOGGER.debug("Incoming Telnet message: %s", message)
        self._last_message_time = time.monotonic()
        if len(message) < 3:
            return

        # Event is 2 characters
        event = self._get_event(message)
        # Parameter is the remaining characters
        parameter = message[len(event) :]

        if event == "MV":
            # This seems undocumented by Denon and appears to basically be a
            # noop that goes along with volume changes. This is here to prevent
            # duplicate callback calls.
            if parameter[0:3] == "MAX":
                return

        # Determine zone
        zone = MAIN_ZONE
        if event in ALL_ZONE_TELNET_EVENTS:
            zone = ALL_ZONES
        elif event in {"Z2", "Z3"}:
            if event == "Z2":
                zone = ZONE2
            else:
                zone = ZONE3

            if parameter in TELNET_SOURCES:
                event = "SI"
            elif parameter.isdigit():
                event = "MV"
            elif self._get_event(parameter):
                event = self._get_event(parameter)
                parameter = parameter[len(event) :]

        if event not in TELNET_EVENTS:
            return

        task = asyncio.create_task(
            self._async_run_callbacks(message, event, zone, parameter)
        )
        self._callback_tasks.add(task)
        task.add_done_callback(self._callback_tasks.discard)

    async def _async_run_callbacks(
        self, message: str, event: str, zone: str, parameter: str
    ) -> None:
        """Handle triggering the registered callbacks."""
        for callback in self._raw_callbacks:
            try:
                await callback(message)
            except Exception as err:  # pylint: disable=broad-except
                # We don't want a single bad callback to trip up the
                # whole system and prevent further execution
                _LOGGER.error(
                    "%s: Raw callback caused an unhandled exception: %s",
                    self.host,
                    err,
                )

        if event in self._callbacks.keys():
            for callback in self._callbacks[event]:
                try:
                    await callback(zone, event, parameter)
                except Exception as err:  # pylint: disable=broad-except
                    # We don't want a single bad callback to trip up the
                    # whole system and prevent further execution
                    _LOGGER.error(
                        "%s: Event callback caused an unhandled exception: %s",
                        self.host,
                        err,
                    )

        if ALL_TELNET_EVENTS in self._callbacks.keys():
            for callback in self._callbacks[ALL_TELNET_EVENTS]:
                try:
                    await callback(zone, event, parameter)
                except Exception as err:  # pylint: disable=broad-except
                    # We don't want a single bad callback to trip up the
                    # whole system and prevent further execution
                    _LOGGER.error(
                        "%s: Event callback caused an unhandled exception: %s",
                        self.host,
                        err,
                    )

    def _get_event(self, message: str) -> str:
        """Get event of a telnet message."""
        events = self._telnet_event_map.get(message[0:2], [""])
        for event in events:
            if message.startswith(event):
                return event
        return ""

    async def _async_send_confirmation_callback(self, message: str) -> None:
        """Confirm that the telnet command has been executed."""
        if len(message) < 3:
            return
        command = self._send_confirmation_command
        if self._get_event(message) == self._get_event(self._send_confirmation_command):
            self._send_confirmation_command = ""
            self._send_confirmation_event.set()
            _LOGGER.debug("Command %s confirmed", command)

    async def _async_send_command(
        self, command: str, skip_confirmation: bool = False
    ) -> None:
        """Send one telnet command to the receiver."""
        async with self._send_lock:
            if not skip_confirmation:
                self._send_confirmation_command = command
                self._send_confirmation_event.clear()
            if not self.connected or not self.healthy:
                raise AvrProcessingError(
                    f"Error sending command {command}. Telnet connected: "
                    f"{self.connected}, Connection healthy: {self.healthy}"
                )
            self._protocol.write(f"{command}\r")
            if not skip_confirmation:
                try:
                    await asyncio.wait_for(
                        self._send_confirmation_event.wait(),
                        self._send_confirmation_timeout,
                    )
                except asyncio.TimeoutError:
                    _LOGGER.info(
                        "Timeout waiting for confirmation of command: %s", command
                    )
                finally:
                    self._send_confirmation_command = ""

    async def async_send_commands(
        self, *commands: str, skip_confirmation: bool = False
    ) -> None:
        """Send telnet commands to the receiver."""
        for command in commands:
            await self._async_send_command(command, skip_confirmation=skip_confirmation)

    def send_commands(self, *commands: str, skip_confirmation: bool = False) -> None:
        """Send telnet commands to the receiver."""
        task = asyncio.create_task(
            self.async_send_commands(*commands, skip_confirmation=skip_confirmation)
        )
        self._send_tasks.add(task)
        task.add_done_callback(self._send_tasks.discard)

    ##############
    # Properties #
    ##############
    @property
    def connected(self) -> bool:
        """Return True if telnet connection is enabled."""
        return self._connection_enabled

    @property
    def healthy(self) -> bool:
        """Return True if telnet connection is healthy."""
        return self._protocol is not None and self._protocol.connected



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the interface to Denon AVR receivers.

:copyright: (c) 2016 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import asyncio
import logging
import time
from typing import Awaitable, Callable, Dict, List, Literal, Optional, Union

import attr
import httpx

from .audyssey import DenonAVRAudyssey, audyssey_factory
from .const import (
    DENON_ATTR_SETATTR,
    MAIN_ZONE,
    VALID_ZONES,
    AudioRestorers,
    AutoStandbys,
    BluetoothOutputModes,
    DimmerModes,
    EcoModes,
    HDMIAudioDecodes,
    HDMIOutputs,
    PanelLocks,
    RoomSizes,
    TransducerLPFs,
    VideoProcessingModes,
)
from .dirac import DenonAVRDirac, dirac_factory
from .exceptions import AvrCommandError
from .foundation import DenonAVRFoundation, set_api_host, set_api_timeout
from .input import DenonAVRInput, input_factory
from .soundmode import DenonAVRSoundMode, sound_mode_factory
from .tonecontrol import DenonAVRToneControl, tone_control_factory
from .volume import DenonAVRVolume, volume_factory

_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVR(DenonAVRFoundation):
    """
    Representing a Denon AVR Device.

    Initialize MainZone of DenonAVR.

    :param host: IP or HOSTNAME.
    :type host: str

    :param name: Device name, if None FriendlyName of device is used.
    :type name: str or None

    :param show_all_inputs: If True deleted input functions are also shown
    :type show_all_inputs: bool

    :param timeout: Timeout when calling device APIs.
    :type timeout: float

    :param add_zones: Additional Zones for which an instance are created
    :type add_zones: dict [str, str] or None
    """

    _host: str = attr.ib(converter=str, on_setattr=[*DENON_ATTR_SETATTR, set_api_host])
    _name: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _show_all_inputs: bool = attr.ib(converter=bool, default=False)
    _add_zones: Optional[Dict[str, str]] = attr.ib(
        validator=attr.validators.optional(
            attr.validators.deep_mapping(
                attr.validators.in_(VALID_ZONES),
                attr.validators.optional(attr.validators.instance_of(str)),
                attr.validators.instance_of(dict),
            )
        ),
        default=None,
    )
    _timeout: float = attr.ib(
        converter=float, on_setattr=[*DENON_ATTR_SETATTR, set_api_timeout], default=2.0
    )
    _zones: Dict[str, DenonAVRFoundation] = attr.ib(
        validator=attr.validators.deep_mapping(
            attr.validators.in_(VALID_ZONES),
            attr.validators.instance_of(DenonAVRFoundation),
            attr.validators.instance_of(dict),
        ),
        default=attr.Factory(dict),
        init=False,
    )
    _setup_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
    audyssey: DenonAVRAudyssey = attr.ib(
        validator=attr.validators.instance_of(DenonAVRAudyssey),
        default=attr.Factory(audyssey_factory, takes_self=True),
        init=False,
    )
    dirac: DenonAVRDirac = attr.ib(
        validator=attr.validators.instance_of(DenonAVRDirac),
        default=attr.Factory(dirac_factory, takes_self=True),
        init=False,
    )
    input: DenonAVRInput = attr.ib(
        validator=attr.validators.instance_of(DenonAVRInput),
        default=attr.Factory(input_factory, takes_self=True),
        init=False,
    )
    soundmode: DenonAVRSoundMode = attr.ib(
        validator=attr.validators.instance_of(DenonAVRSoundMode),
        default=attr.Factory(sound_mode_factory, takes_self=True),
        init=False,
    )
    tonecontrol: DenonAVRToneControl = attr.ib(
        validator=attr.validators.instance_of(DenonAVRToneControl),
        default=attr.Factory(tone_control_factory, takes_self=True),
        init=False,
    )
    vol: DenonAVRVolume = attr.ib(
        validator=attr.validators.instance_of(DenonAVRVolume),
        default=attr.Factory(volume_factory, takes_self=True),
        init=False,
    )

    def __attrs_post_init__(self) -> None:
        """Initialize special attributes."""
        # Set host and timeout again to start its custom setattr function
        self._host = self._host
        self._timeout = self._timeout

        # Add own instance to zone dictionary
        self._zones[self._device.zone] = self

        # Create instances of additional zones if requested
        if self._device.zone == MAIN_ZONE and self._add_zones is not None:
            self.create_zones(self._add_zones)

    def create_zones(self, add_zones):
        """Create instances of additional zones for the receiver."""
        for zone, zname in add_zones.items():
            # Name either set explicitly or name of Main Zone with suffix
            zonename = None
            if zname is None and self._name is not None:
                zonename = f"{self._name} {zone}"
            zone_device = attr.evolve(self._device, zone=zone)
            zone_inst = DenonAVR(
                host=self._host,
                device=zone_device,
                name=zonename,
                timeout=self._timeout,
                show_all_inputs=self._show_all_inputs,
            )
            self._zones[zone] = zone_inst

    async def async_setup(self) -> None:
        """Ensure that configuration is loaded from receiver asynchronously."""
        async with self._setup_lock:
            _LOGGER.debug("Starting denonavr setup")
            # Device setup
            await self._device.async_setup()
            if self._name is None:
                self._name = self._device.friendly_name

            # Setup other functions
            self.input.setup()
            await self.soundmode.async_setup()
            await self.tonecontrol.async_setup()
            self.vol.setup()
            self.audyssey.setup()
            self.dirac.setup()

            for zone_name, zone_item in self._zones.items():
                if zone_name != self.zone:
                    await zone_item.async_setup()

            self._is_setup = True
            _LOGGER.debug("Finished denonavr setup")

    async def async_update(self):
        """
        Get the latest status information from device asynchronously.

        Method executes the update method for the current receiver type.
        """
        _LOGGER.debug("Starting denonavr update")
        # Ensure that the device is setup
        if not self._is_setup:
            await self.async_setup()

        # Create a cache id for this global update
        cache_id = time.time()

        # Verify update method
        _LOGGER.debug("Verifying update method")
        await self._device.async_verify_avr_2016_update_method(cache_id=cache_id)

        # Update device
        await self._device.async_update(global_update=True, cache_id=cache_id)

        # Update other functions
        await self.input.async_update(global_update=True, cache_id=cache_id)
        await self.soundmode.async_update(global_update=True, cache_id=cache_id)
        await self.tonecontrol.async_update(global_update=True, cache_id=cache_id)
        await self.vol.async_update(global_update=True, cache_id=cache_id)

        # AppCommand0300.xml interface is very slow, thus it is not included
        # into main update
        # await self.audyssey.async_update(
        #     global_update=True, cache_id=cache_id)
        _LOGGER.debug("Finished denonavr update")

    async def async_update_tonecontrol(self):
        """Get Tonecontrol settings."""
        await self.tonecontrol.async_update()

    async def async_update_audyssey(self):
        """Get Audyssey settings."""
        await self.audyssey.async_update()

    async def async_get_command(self, request: str) -> str:
        """Send HTTP GET command to Denon AVR receiver asynchronously."""
        return await self._device.api.async_get_command(request)

    async def async_send_telnet_commands(self, *commands: str) -> None:
        """Send telnet commands to the receiver."""
        await self._device.telnet_api.async_send_commands(*commands)

    def send_telnet_commands(self, *commands: str) -> None:
        """Send telnet commands to the receiver."""
        self._device.telnet_api.send_commands(*commands)

    def register_callback(
        self, event: str, callback: Callable[[str, str, str], Awaitable[None]]
    ):
        """Register a callback for telnet events."""
        self._device.telnet_api.register_callback(event, callback=callback)

    def unregister_callback(
        self, event: str, callback: Callable[[str, str, str], Awaitable[None]]
    ):
        """Unregister a callback for telnet events."""
        self._device.telnet_api.unregister_callback(event, callback=callback)

    async def async_telnet_connect(self):
        """Connect to the telnet interface of the receiver."""
        await self._device.telnet_api.async_connect()

    async def async_telnet_disconnect(self):
        """Disconnect from the telnet interface of the receiver."""
        await self._device.telnet_api.async_disconnect()

    ##############
    # Properties #
    ##############
    @property
    def zone(self) -> str:
        """Return Zone of this instance."""
        return self._device.zone

    @property
    def zones(self) -> Dict[str, DenonAVRFoundation]:
        """Return all Zone instances of the device."""
        zones = self._zones.copy()
        return zones

    @property
    def name(self) -> Optional[str]:
        """Return the name of the device as string."""
        return self._name

    @property
    def host(self) -> str:
        """Return the host of the device as string."""
        return self._host

    @property
    def manufacturer(self) -> Optional[str]:
        """Return the manufacturer of the device as string."""
        return self._device.manufacturer

    @property
    def model_name(self) -> Optional[str]:
        """Return the model name of the device as string."""
        return self._device.model_name

    @property
    def serial_number(self) -> Optional[str]:
        """Return the serial number of the device as string."""
        return self._device.serial_number

    @property
    def power(self) -> Optional[str]:
        """
        Return the power state of the device.

        Possible values are: "ON", "STANDBY" and "OFF"
        """
        return self._device.power

    @property
    def settings_menu(self) -> Optional[bool]:
        """
        Return the settings menu state of the device.

        Only available if using Telnet.
        """
        return self._device.settings_menu

    @property
    def state(self) -> Optional[str]:
        """
        Return the state of the device.

        Possible values are: "on", "off", "playing", "paused"
        "playing" and "paused" are only available for input functions
        in PLAYING_SOURCES.
        """
        return self.input.state

    @property
    def muted(self) -> bool:
        """
        Boolean if volume is currently muted.

        Return "True" if muted and "False" if not muted.
        """
        return self.vol.muted

    @property
    def volume(self) -> float:
        """
        Return volume of Denon AVR as float.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        return self.vol.volume

    @property
    def input_func(self) -> Optional[str]:
        """Return the current input source as string."""
        return self.input.input_func

    @property
    def input_func_list(self) -> List[str]:
        """Return a list of available input sources as string."""
        return self.input.input_func_list

    @property
    def support_sound_mode(self) -> Optional[bool]:
        """Return True if sound mode is supported."""
        return self.soundmode.support_sound_mode

    @property
    def sound_mode(self) -> Optional[str]:
        """Return the matched current sound mode as a string."""
        return self.soundmode.sound_mode

    @property
    def sound_mode_list(self) -> List[str]:
        """Return a list of available sound modes as string."""
        return self.soundmode.sound_mode_list

    @property
    def sound_mode_map(self) -> Dict[str, str]:
        """Return a dict of available sound modes with their mapping values."""
        return self.soundmode.sound_mode_map

    @property
    def sound_mode_map_rev(self) -> Dict[str, str]:
        """Return a dict to map each sound_mode_raw to matching sound_mode."""
        return self.soundmode.sound_mode_map_rev

    @property
    def sound_mode_raw(self) -> Optional[str]:
        """Return the current sound mode as string as received from the AVR."""
        return self.soundmode.sound_mode_raw

    @property
    def image_url(self) -> Optional[str]:
        """Return image URL of current playing media when powered on."""
        return self.input.image_url

    @property
    def title(self) -> Optional[str]:
        """Return title of current playing media as string."""
        return self.input.title

    @property
    def artist(self) -> Optional[str]:
        """Return artist of current playing media as string."""
        return self.input.artist

    @property
    def album(self) -> Optional[str]:
        """Return album name of current playing media as string."""
        return self.input.album

    @property
    def band(self) -> Optional[str]:
        """Return band of current radio station as string."""
        return self.input.band

    @property
    def frequency(self) -> Optional[str]:
        """Return frequency of current radio station as string."""
        return self.input.frequency

    @property
    def station(self) -> Optional[str]:
        """Return current radio station as string."""
        return self.input.station

    @property
    def netaudio_func_list(self) -> List[str]:
        """Return list of network audio devices.

        Those devices should react to play, pause, next and previous
        track commands.
        """
        return self.input.netaudio_func_list

    @property
    def playing_func_list(self) -> List[str]:
        """Return list of playing devices.

        Those devices offer additional information about what they are playing
        (e.g. title, artist, album, band, frequency, station, image_url).
        """
        return self.input.playing_func_list

    @property
    def receiver_port(self) -> int:
        """Return the receiver's port."""
        if self._device.receiver is None:
            return None
        return self._device.receiver.port

    @property
    def receiver_type(self) -> Optional[str]:
        """Return the receiver's type."""
        if self._device.receiver is None:
            return None
        return self._device.receiver.type

    @property
    def telnet_available(self) -> bool:
        """Return True if telnet is connected and healthy."""
        return self._device.telnet_available

    @property
    def telnet_connected(self) -> bool:
        """Return True if telnet is connected."""
        return self._device.telnet_api.connected

    @property
    def telnet_healthy(self) -> bool:
        """Return True if telnet connection is healthy."""
        return self._device.telnet_api.healthy

    @property
    def show_all_inputs(self) -> Optional[bool]:
        """Indicate if all inputs are shown or just active one."""
        return self._show_all_inputs

    @property
    def support_tone_control(self) -> Optional[bool]:
        """Return True if tone control is supported."""
        return self.tonecontrol.support_tone_control

    @property
    def tone_control_status(self) -> Optional[bool]:
        """Return value of tone control status."""
        return self.tonecontrol.tone_control_status

    @property
    def tone_control_adjust(self) -> Optional[bool]:
        """Return value of tone control adjust."""
        return self.tonecontrol.tone_control_adjust

    @property
    def bass(self) -> Optional[int]:
        """Return value of bass."""
        return self.tonecontrol.bass

    @property
    def bass_level(self) -> Optional[str]:
        """Return level of bass."""
        return self.tonecontrol.bass_level

    @property
    def treble(self) -> Optional[int]:
        """Return value of treble."""
        return self.tonecontrol.treble

    @property
    def treble_level(self) -> Optional[str]:
        """Return level of treble."""
        return self.tonecontrol.treble_level

    @property
    def dynamic_eq(self) -> Optional[bool]:
        """Return value of Dynamic EQ."""
        return self.audyssey.dynamic_eq

    @property
    def reference_level_offset(self) -> Optional[str]:
        """Return value of Reference Level Offset."""
        return self.audyssey.reference_level_offset

    @property
    def reference_level_offset_setting_list(self) -> List[str]:
        """Return a list of available reference level offset settings."""
        return self.audyssey.reference_level_offset_setting_list

    @property
    def dynamic_volume(self) -> Optional[str]:
        """Return value of Dynamic Volume."""
        return self.audyssey.dynamic_volume

    @property
    def dynamic_volume_setting_list(self) -> List[str]:
        """Return a list of available Dynamic Volume settings."""
        return self.audyssey.dynamic_volume_setting_list

    @property
    def multi_eq(self) -> Optional[str]:
        """Return value of MultiEQ."""
        return self.audyssey.multi_eq

    @property
    def multi_eq_setting_list(self) -> List[str]:
        """Return a list of available MultiEQ settings."""
        return self.audyssey.multi_eq_setting_list

    @property
    def dimmer(self) -> Optional[str]:
        """
        Returns the dimmer state of the device.

        Only available if using Telnet.

        Possible values are: "Off", "Dark", "Dim" and "Bright"
        """
        return self._device.dimmer

    @property
    def auto_standby(self) -> Optional[str]:
        """
        Return the auto-standby state of the device.

        Only available if using Telnet.

        Possible values are: "OFF", "15M", "30M", "60M"
        """
        return self._device.auto_standby

    @property
    def sleep(self) -> Optional[Union[str, int]]:
        """
        Return the sleep timer for the device.

        Only available if using Telnet.

        Possible values are: "OFF" and 1-120 (in minutes)
        """
        return self._device.sleep

    @property
    def delay(self) -> Optional[int]:
        """
        Return the audio delay for the device in ms.

        Only available if using Telnet.
        """
        return self._device.delay

    @property
    def eco_mode(self) -> Optional[str]:
        """
        Returns the eco-mode for the device.

        Only available if using Telnet.

        Possible values are: "Off", "On", "Auto"
        """
        return self._device.eco_mode

    @property
    def hdmi_output(self) -> Optional[str]:
        """
        Returns the HDMI-output for the device.

        Only available if using Telnet.

        Possible values are: "Auto", "HDMI1", "HDMI2"
        """
        return self._device.hdmi_output

    @property
    def hdmi_audio_decode(self) -> Optional[str]:
        """
        Returns the HDMI Audio Decode mode for the device.

        Only available if using Telnet.

        Possible values are: "AMP", "TV"
        """
        return self._device.hdmi_audio_decode

    @property
    def video_processing_mode(self) -> Optional[str]:
        """
        Return the video processing mode for the device.

        Only available if using Telnet.

        Possible values are: "Auto", "Game", "Movie", "Bypass"
        """
        return self._device.video_processing_mode

    @property
    def tactile_transducer(self) -> Optional[bool]:
        """
        Return the tactile transducer state of the device.

        Only available if using Telnet.
        """
        return self._device.tactile_transducer

    @property
    def tactile_transducer_level(self) -> Optional[float]:
        """
        Return the tactile transducer level in dB.

        Only available if using Telnet.
        """
        return self._device.tactile_transducer_level

    @property
    def tactile_transducer_lpf(self) -> Optional[str]:
        """
        Return the tactile transducer low pass filter frequency.

        Only available if using Telnet.
        """
        return self._device.tactile_transducer_lpf

    @property
    def room_size(self) -> Optional[str]:
        """
        Return the room size for the device.

        Only available if using Telnet.

        Possible values are: "S", "MS", "M", "ML", "L"
        """
        return self._device.room_size

    @property
    def triggers(self) -> Dict[int, str]:
        """
        Return the triggers and their statuses for the device.

        Only available if using Telnet.
        """
        return self._device.triggers

    @property
    def speaker_preset(self) -> Optional[int]:
        """
        Return the speaker preset for the device.

        Only available if using Telnet.

        Possible values are: "1", "2"
        """
        return self._device.speaker_preset

    @property
    def bt_transmitter(self) -> Optional[bool]:
        """
        Return the Bluetooth transmitter state for the device.

        Only available if using Telnet.
        """
        return self._device.bt_transmitter

    @property
    def bt_output_mode(self) -> Optional[str]:
        """
        Return the Bluetooth output mode for the device.

        Only available if using Telnet.

        Possible values are: "Bluetooth + Speakers", "Bluetooth Only"
        """
        return self._device.bt_output_mode

    @property
    def delay_time(self) -> Optional[int]:
        """
        Return the delay time for the device in ms.

        Only available if using Telnet.
        """
        return self._device.delay_time

    @property
    def audio_restorer(self) -> Optional[str]:
        """
        Return the audio restorer for the device.

        Only available if using Telnet.

        Possible values are: "Off", "Low", "Medium", "High"
        """
        return self._device.audio_restorer

    @property
    def graphic_eq(self) -> Optional[bool]:
        """
        Return the Graphic EQ status for the device.

        Only available if using Telnet.
        """
        return self._device.graphic_eq

    @property
    def headphone_eq(self) -> Optional[bool]:
        """
        Return the Headphone EQ status for the device.

        Only available if using Telnet.
        """
        return self._device.headphone_eq

    ##########
    # Getter #
    ##########

    def get_trigger(self, trigger: int) -> Optional[str]:
        """
        Return the status of a specific trigger.

        Only available if using Telnet.

        Valid trigger values are 1-3.
        """
        return self._device.get_trigger(trigger)

    ##########
    # Setter #
    ##########
    def set_async_client_getter(
        self, async_client_getter: Callable[[], httpx.AsyncClient]
    ) -> None:
        """
        Set a custom httpx.AsyncClient getter for this instance.

        The function provided must return an instance of httpx.AsyncClient.
        This is a non-blocking method.
        """
        if not callable(async_client_getter):
            raise AvrCommandError("Provided object is not callable")
        self._device.api.httpx_async_client.client_getter = async_client_getter

    async def async_dynamic_eq_off(self) -> None:
        """Turn DynamicEQ off."""
        await self.audyssey.async_dynamiceq_off()

    async def async_dynamic_eq_on(self) -> None:
        """Turn DynamicEQ on."""
        await self.audyssey.async_dynamiceq_on()

    async def async_toggle_dynamic_eq(self) -> None:
        """Toggle DynamicEQ."""
        await self.audyssey.async_toggle_dynamic_eq()

    async def async_set_multieq(self, value: str) -> None:
        """Set MultiEQ mode."""
        await self.audyssey.async_set_multieq(value)

    async def async_set_reflevoffset(self, value: str) -> None:
        """Set Reference Level Offset."""
        await self.audyssey.async_set_reflevoffset(value)

    async def async_set_dynamicvol(self, value: str) -> None:
        """Set Dynamic Volume."""
        await self.audyssey.async_set_dynamicvol(value)

    async def async_set_input_func(self, input_func: str) -> None:
        """
        Set input_func of device.

        Valid values depend on the device and should be taken from
        "input_func_list".
        """
        await self.input.async_set_input_func(input_func)

    async def async_set_sound_mode(self, sound_mode: str) -> None:
        """
        Set sound_mode of device.

        Valid values depend on the device and should be taken from
        "sound_mode_list".
        """
        await self.soundmode.async_set_sound_mode(sound_mode)

    async def async_toggle_play_pause(self) -> None:
        """Toggle play pause media player."""
        await self.input.async_toggle_play_pause()

    async def async_play(self) -> None:
        """Send play command to receiver command via HTTP post."""
        await self.input.async_play()

    async def async_pause(self) -> None:
        """Send pause command to receiver command via HTTP post."""
        await self.input.async_pause()

    async def async_previous_track(self) -> None:
        """Send previous track command to receiver command via HTTP post."""
        await self.input.async_previous_track()

    async def async_next_track(self) -> None:
        """Send next track command to receiver command via HTTP post."""
        await self.input.async_next_track()

    async def async_power_on(self) -> None:
        """Turn on receiver via HTTP get command."""
        await self._device.async_power_on()

    async def async_power_off(self) -> None:
        """Turn off receiver via HTTP get command."""
        await self._device.async_power_off()

    async def async_volume_up(self) -> None:
        """Volume up receiver via HTTP get command."""
        await self.vol.async_volume_up()

    async def async_volume_down(self) -> None:
        """Volume down receiver via HTTP get command."""
        await self.vol.async_volume_down()

    async def async_set_volume(self, volume: float) -> None:
        """
        Set receiver volume via HTTP get command.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        await self.vol.async_set_volume(volume)

    async def async_mute(self, mute: bool) -> None:
        """Mute receiver via HTTP get command."""
        await self.vol.async_mute(mute)

    async def async_enable_tone_control(self) -> None:
        """Enable tone control to change settings like bass or treble."""
        await self.tonecontrol.async_enable_tone_control()

    async def async_disable_tone_control(self) -> None:
        """Disable tone control to change settings like bass or treble."""
        await self.tonecontrol.async_disable_tone_control()

    async def async_set_bass(self, value: int) -> None:
        """
        Set receiver bass.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        await self.tonecontrol.async_set_bass(value)

    async def async_bass_up(self) -> None:
        """
        Increase level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_bass_up()

    async def async_bass_down(self) -> None:
        """
        Decrease level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_bass_down()

    async def async_set_treble(self, value: int) -> None:
        """
        Set receiver treble.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        await self.tonecontrol.async_set_treble(value)

    async def async_treble_up(self) -> None:
        """
        Increase level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_treble_up()

    async def async_treble_down(self) -> None:
        """
        Decrease level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_treble_down()

    async def async_cursor_up(self) -> None:
        """Send cursor up to receiver via HTTP get command."""
        await self._device.async_cursor_up()

    async def async_cursor_down(self) -> None:
        """Send cursor down to receiver via HTTP get command."""
        await self._device.async_cursor_down()

    async def async_cursor_left(self) -> None:
        """Send cursor left to receiver via HTTP get command."""
        await self._device.async_cursor_left()

    async def async_cursor_right(self) -> None:
        """Send cursor right to receiver via HTTP get command."""
        await self._device.async_cursor_right()

    async def async_cursor_enter(self) -> None:
        """Send cursor enter to receiver via HTTP get command."""
        await self._device.async_cursor_enter()

    async def async_back(self) -> None:
        """Send back to receiver via HTTP get command."""
        await self._device.async_back()

    async def async_info(self) -> None:
        """Send info to receiver via HTTP get command."""
        await self._device.async_info()

    async def async_options(self) -> None:
        """Raise options menu to receiver via HTTP get command."""
        await self._device.async_options()

    async def async_settings_menu(self) -> None:
        """Raise settings menu to receiver via HTTP get command."""
        await self._device.async_settings_menu()

    async def async_channel_level_adjust(self) -> None:
        """Toggle the channel level adjust menu on receiver via HTTP get command."""
        await self._device.async_channel_level_adjust()

    async def async_dimmer_toggle(self) -> None:
        """Toggle dimmer on receiver via HTTP get command."""
        await self._device.async_dimmer_toggle()

    async def async_dimmer(self, mode: DimmerModes) -> None:
        """Set dimmer mode on receiver via HTTP get command."""
        await self._device.async_dimmer(mode)

    async def async_auto_standby(self, auto_standby: AutoStandbys) -> None:
        """Set auto standby on receiver via HTTP get command."""
        await self._device.async_auto_standby(auto_standby)

    async def async_sleep(self, sleep: Union[Literal["OFF"], int]) -> None:
        """
        Set auto standby on receiver via HTTP get command.

        Valid sleep values are "OFF" and 1-120 (in minutes)
        """
        await self._device.async_sleep(sleep)

    async def async_delay_up(self) -> None:
        """Increase delay of the audio."""
        await self._device.async_delay_up()

    async def async_delay_down(self) -> None:
        """Decrease delay of the audio."""
        await self._device.async_delay_down()

    async def async_eco_mode(self, mode: EcoModes) -> None:
        """Set Eco mode."""
        await self._device.async_eco_mode(mode)

    async def async_hdmi_output(self, output: HDMIOutputs) -> None:
        """Set HDMI output."""
        await self._device.async_hdmi_output(output)

    async def async_hdmi_audio_decode(self, mode: HDMIAudioDecodes) -> None:
        """Set HDMI Audio Decode mode on receiver via HTTP get command."""
        await self._device.async_hdmi_audio_decode(mode)

    async def async_video_processing_mode(self, mode: VideoProcessingModes) -> None:
        """Set video processing mode on receiver via HTTP get command."""
        await self._device.async_video_processing_mode(mode)

    async def async_status(self) -> str:
        """
        Toggles the display of status on the device.

        Only supported on Denon models.
        """
        return await self._device.async_status()

    async def async_system_reset(self) -> None:
        """DANGER! Reset the receiver via HTTP get command."""
        await self._device.async_system_reset()

    async def async_network_restart(self) -> None:
        """Restart the network on the receiver via HTTP get command."""
        await self._device.async_network_restart()

    async def async_speaker_preset(self, preset: int) -> None:
        """
        Set speaker preset on receiver via HTTP get command.

        Valid preset values are 1-2.
        """
        await self._device.async_speaker_preset(preset)

    async def async_speaker_preset_toggle(self) -> None:
        """
        Toggle speaker preset on receiver via HTTP get command.

        Only available if using Telnet.
        """
        await self._device.async_speaker_preset_toggle()

    async def async_bt_transmitter_on(self) -> None:
        """Turn on Bluetooth transmitter on receiver via HTTP get command."""
        await self._device.async_bt_transmitter_on()

    async def async_bt_transmitter_off(self) -> None:
        """Turn off Bluetooth transmitter on receiver via HTTP get command."""
        await self._device.async_bt_transmitter_off()

    async def async_bt_transmitter_toggle(self) -> None:
        """
        Toggle Bluetooth transmitter mode on receiver via HTTP get command.

        Only available if using Telnet.
        """
        await self._device.async_bt_transmitter_toggle()

    async def async_bt_output_mode(self, mode: BluetoothOutputModes) -> None:
        """Set Bluetooth transmitter mode on receiver via HTTP get command."""
        await self._device.async_bt_output_mode(mode)

    async def async_bt_output_mode_toggle(self) -> None:
        """
        Toggle Bluetooth output mode on receiver via HTTP get command.

        Only available if using Telnet.
        """
        await self._device.async_bt_output_mode_toggle()

    async def async_delay_time_up(self) -> None:
        """Delay time up on receiver via HTTP get command."""
        await self._device.async_delay_time_up()

    async def async_delay_time_down(self) -> None:
        """Delay time up on receiver via HTTP get command."""
        await self._device.async_delay_time_down()

    async def async_delay_time(self, delay_time: int) -> None:
        """
        Set delay time on receiver via HTTP get command.

        :param delay_time: Delay time in ms. Valid values are 0-999.
        """
        await self._device.async_delay_time(delay_time)

    async def async_audio_restorer(self, mode: AudioRestorers):
        """Set audio restorer on receiver via HTTP get command."""
        await self._device.async_audio_restorer(mode)

    async def async_remote_control_lock(self):
        """Set remote control lock on receiver via HTTP get command."""
        await self._device.async_remote_control_lock()

    async def async_remote_control_unlock(self):
        """Set remote control unlock on receiver via HTTP get command."""
        await self._device.async_remote_control_unlock()

    async def async_panel_lock(self, panel_lock_mode: PanelLocks):
        """Set panel lock on receiver via HTTP get command."""
        await self._device.async_panel_lock(panel_lock_mode)

    async def async_panel_unlock(self):
        """Set panel unlock on receiver via HTTP get command."""
        await self._device.async_panel_unlock()

    async def async_graphic_eq_on(self) -> None:
        """Turn on Graphic EQ on receiver via HTTP get command."""
        await self._device.async_graphic_eq_on()

    async def async_graphic_eq_off(self) -> None:
        """Turn off Graphic EQ on receiver via HTTP get command."""
        await self._device.async_graphic_eq_off()

    async def async_graphic_eq_toggle(self) -> None:
        """
        Toggle Graphic EQ on receiver via HTTP get command.

        Only available if using Telnet.
        """
        await self._device.async_graphic_eq_toggle()

    async def async_headphone_eq_on(self) -> None:
        """Turn on Headphone EQ on receiver via HTTP get command."""
        await self._device.async_headphone_eq_on()

    async def async_headphone_eq_off(self) -> None:
        """Turn off Headphone EQ on receiver via HTTP get command."""
        await self._device.async_headphone_eq_off()

    async def async_headphone_eq_toggle(self) -> None:
        """
        Toggle Headphone EQ on receiver via HTTP get command.

        Only available if using Telnet.
        """
        await self._device.async_headphone_eq_toggle()

    async def async_tactile_transducer_on(self) -> None:
        """Turn on tactile transducer on receiver via HTTP get command."""
        await self._device.async_tactile_transducer_on()

    async def async_tactile_transducer_off(self) -> None:
        """Turn on tactile transducer on receiver via HTTP get command."""
        await self._device.async_tactile_transducer_off()

    async def async_tactile_transducer_toggle(self) -> None:
        """
        Turn on tactile transducer on receiver via HTTP get command.

        Only available if using Telnet.
        """
        await self._device.async_tactile_transducer_toggle()

    async def async_tactile_transducer_level_up(self) -> None:
        """Increase the transducer level on receiver via HTTP get command."""
        await self._device.async_tactile_transducer_level_up()

    async def async_tactile_transducer_level_down(self) -> None:
        """Decrease the transducer on receiver via HTTP get command."""
        await self._device.async_tactile_transducer_level_down()

    async def async_transducer_lpf(self, lpf: TransducerLPFs) -> None:
        """Set transducer low pass filter on receiver via HTTP get command."""
        await self._device.async_transducer_lpf(lpf)

    async def async_room_size(self, room_size: RoomSizes) -> None:
        """Set room size on receiver via HTTP get command."""
        await self._device.async_room_size(room_size)

    async def async_trigger_on(self, trigger: int) -> None:
        """
        Set trigger to ON on receiver via HTTP get command.

        :param trigger: Trigger number to set to ON. Valid values are 1-3.
        """
        await self._device.async_trigger_on(trigger)

    async def async_trigger_off(self, trigger: int) -> None:
        """
        Set trigger to OFF on receiver via HTTP get command.

        :param trigger: Trigger number to set to OFF. Valid values are 1-3.
        """
        await self._device.async_trigger_off(trigger)

    async def async_trigger_toggle(self, trigger: int) -> None:
        """
        Toggle trigger on receiver via HTTP get command.

        Only available if using Telnet.

        :param trigger: Trigger number to toggle. Valid values are 1-3.
        """
        await self._device.async_trigger_toggle(trigger)

    async def async_quick_select_mode(self, quick_select_number: int) -> None:
        """
        Set quick select mode on receiver via HTTP get command.

        :param quick_select_number: Quick select number to set. Valid values are 1-5.
        """
        await self._device.async_quick_select_mode(quick_select_number)

    async def async_quick_select_memory(self, quick_select_number: int) -> None:
        """
        Set quick select memory on receiver via HTTP get command.

        :param quick_select_number: Quick select number to set. Valid values are 1-5.
        """
        await self._device.async_quick_select_memory(quick_select_number)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module covers some tests for mappings in const.py.

:copyright: (c) 2016 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import unittest
from typing import get_args

from denonavr.const import (
    CHANNEL_MAP,
    DIMMER_MODE_MAP,
    DIRAC_FILTER_MAP,
    ECO_MODE_MAP,
    HDMI_OUTPUT_MAP,
    Channels,
    DimmerModes,
    DiracFilters,
    EcoModes,
    HDMIOutputs,
)


class TestMappings(unittest.TestCase):
    """Test case for mappings in const.py."""

    def test_dimmer_modes_mappings(self):
        """Test that all dimmer modes are in the mapping."""
        for mode in get_args(DimmerModes):
            self.assertIn(mode, DIMMER_MODE_MAP)

    def test_eco_modes_mappings(self):
        """Test that all eco-modes are in the mapping."""
        for mode in get_args(EcoModes):
            self.assertIn(mode, ECO_MODE_MAP)

    def test_hdmi_outputs_mappings(self):
        """Test that all hdmi outputs are in the mapping."""
        for hdmi_output in get_args(HDMIOutputs):
            self.assertIn(hdmi_output, HDMI_OUTPUT_MAP)

    def test_channel_mappings(self):
        """Test that all channels are in the mapping."""
        for channel in get_args(Channels):
            self.assertIn(channel, CHANNEL_MAP)

    def test_dirac_filter_mappings(self):
        """Test that dirac filters are in the mapping."""
        for dirac_filter in get_args(DiracFilters):
            self.assertIn(dirac_filter, DIRAC_FILTER_MAP)


if __name__ == "__main__":
    unittest.main()
